#!/usr/bin/env python
# -*- coding: utf-8
from __future__ import unicode_literals
from __future__ import print_function

import os
import click
import traceback

from types import GeneratorType
from prompt_toolkit import (AbortAction,
                            Application,
                            CommandLineInterface)
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Always, HasFocus, IsDone
from prompt_toolkit.layout.processors import (
    HighlightMatchingBracketProcessor,
    ConditionalProcessor
)
from prompt_toolkit.buffer import (Buffer, AcceptAction)
from prompt_toolkit.shortcuts import (create_prompt_layout, create_eventloop)
from prompt_toolkit.history import FileHistory

from .client import DockerClient
from .client import DockerPermissionException
from .client import DockerTimeoutException
from .client import DockerSslException
from .completer import DockerCompleter
from .lexer import CommandLexer
from .formatter import format_data
from .formatter import output_stream
from .config import write_default_config, read_config
from .style import style_factory
from .keys import get_key_manager
from .toolbar import create_toolbar_handler
from .options import OptionError
from .logger import create_logger
from .__init__ import __version__

# Added this to avoid the annoying warning: http://click.pocoo.org/5/python3/
click.disable_unicode_literals_warning = True


class WharfeeCli(object):
    """
    The CLI implementation.
    """

    dcli = None
    keyword_completer = None
    handler = None
    saved_less_opts = None
    config = None
    config_template = 'wharfeerc'
    config_name = '~/.wharfeerc'

    def __init__(self, no_completion=False):
        """
        Initialize class members.
        Should read the config here at some point.
        """

        self.config = self.read_configuration()
        self.theme = self.config['main']['theme']

        log_file = self.config['main']['log_file']
        log_level = self.config['main']['log_level']
        self.logger = create_logger(__name__, log_file, log_level)

        # set_completer_options refreshes all by default
        self.handler = DockerClient(
            self.config['main'].as_int('client_timeout'),
            self.clear,
            self.refresh_completions_force,
            self.logger)

        self.completer = DockerCompleter(
            long_option_names=self.get_long_options(),
            fuzzy=self.get_fuzzy_match())
        self.set_completer_options()
        self.completer.set_enabled(not no_completion)
        self.saved_less_opts = self.set_less_opts()

    def read_configuration(self):
        """

        :return:
        """
        default_config = os.path.join(
            self.get_package_path(), self.config_template)
        write_default_config(default_config, self.config_name)
        return read_config(self.config_name, default_config)

    def get_package_path(self):
        """
        Find out pakage root path.
        :return: string: path
        """
        from wharfee import __file__ as package_root
        return os.path.dirname(package_root)

    def set_less_opts(self):
        """
        Set the "less" options and save the old settings.

        What we're setting:
          -F:
            --quit-if-one-screen: Quit if entire file fits on first screen.
          -R:
            --raw-control-chars: Output "raw" control characters.
          -X:
            --no-init: Don't use termcap keypad init/deinit strings.
            --no-keypad: Don't use termcap init/deinit strings.
            This also disables launching "less" in an alternate screen.

        :return: string with old options
        """
        opts = os.environ.get('LESS', '')
        os.environ['LESS'] = '-RXF'
        return opts

    def revert_less_opts(self):
        """
        Restore the previous "less" options.
        """
        os.environ['LESS'] = self.saved_less_opts

    def write_config_file(self):
        """
        Write config file on exit.
        """
        self.config.write()

    def clear(self):
        """
        Clear the screen.
        """
        click.clear()

    def set_completer_options(self, cons=True, runs=True, imgs=True, vols=True):
        """
        Set image and container names in Completer.
        Re-read if needed after a command.
        :param cons: boolean: need to refresh containers
        :param runs: boolean: need to refresh running containers
        :param imgs: boolean: need to refresh images
        :param vols: boolean: need to refresh volumes
        """

        if cons:
            cs = self.handler.containers(all=True)
            if cs and len(cs) > 0 and isinstance(cs[0], dict):
                containers = [name for c in cs for name in c['Names']]
                self.completer.set_containers(containers)

        if runs:
            cs = self.handler.containers()
            if cs and len(cs) > 0 and isinstance(cs[0], dict):
                running = [name for c in cs for name in c['Names']]
                self.completer.set_running(running)

        if imgs:
            def format_tagged(tagname, img_id):
                if tagname == '<none>:<none>':
                    return img_id[:11]
                return tagname

            def parse_image_name(tag, img_id):
                if ':' in tag:
                    result = tag.split(':', 2)[0]
                else:
                    result = tag
                if result == '<none>':
                    result = img_id[:11]
                return result

            ims = self.handler.images()
            if ims and len(ims) > 0 and isinstance(ims[0], dict):
                images = set([])
                tagged = set([])
                for im in ims:
                    repo_tag = '{0}:{1}'.format(im['Repository'], im['Tag'])
                    images.add(parse_image_name(repo_tag, im['Id']))
                    tagged.add(format_tagged(repo_tag, im['Id']))
                self.completer.set_images(images)
                self.completer.set_tagged(tagged)

        if vols:
            vs = self.handler.volume_ls(quiet=True)
            self.completer.set_volumes(vs)

    def set_fuzzy_match(self, is_fuzzy):
        """
        Setter for fuzzy matching mode
        :param is_fuzzy: boolean
        """
        self.config['main']['fuzzy_match'] = is_fuzzy
        self.completer.set_fuzzy_match(is_fuzzy)

    def get_fuzzy_match(self):
        """
        Getter for fuzzy matching mode
        :return: boolean
        """
        return self.config['main'].as_bool('fuzzy_match')

    def set_long_options(self, is_long):
        """
        Setter for long option names.
        :param is_long: boolean
        """
        self.config['main']['suggest_long_option_names'] = is_long
        self.completer.set_long_options(is_long)

    def get_long_options(self):
        """
        Getter for long option names.
        :return: boolean
        """
        return self.config['main'].as_bool('suggest_long_option_names')

    def refresh_completions_force(self):
        """Force refresh and make it visible."""
        self.set_completer_options()
        click.echo('Refreshed completions.')

    def refresh_completions(self):
        """
        After processing the command, refresh the lists of
        containers and images as needed
        """
        self.set_completer_options(self.handler.is_refresh_containers,
                                   self.handler.is_refresh_running,
                                   self.handler.is_refresh_images,
                                   self.handler.is_refresh_volumes)

    def run_cli(self):
        """
        Run the main loop
        """
        print('Version:', __version__)
        print('Home: http://wharfee.com')

        history = FileHistory(os.path.expanduser('~/.wharfee-history'))
        toolbar_handler = create_toolbar_handler(self.get_long_options, self.get_fuzzy_match)

        layout = create_prompt_layout(
            message='wharfee> ',
            lexer=CommandLexer,
            get_bottom_toolbar_tokens=toolbar_handler,
            extra_input_processors=[
                ConditionalProcessor(
                    processor=HighlightMatchingBracketProcessor(
                        chars='[](){}'),
                    filter=HasFocus(DEFAULT_BUFFER) & ~IsDone())
            ]
        )

        cli_buffer = Buffer(
            history=history,
            completer=self.completer,
            complete_while_typing=Always(),
            accept_action=AcceptAction.RETURN_DOCUMENT)

        manager = get_key_manager(
            self.set_long_options,
            self.get_long_options,
            self.set_fuzzy_match,
            self.get_fuzzy_match)

        application = Application(
            style=style_factory(self.theme),
            layout=layout,
            buffer=cli_buffer,
            key_bindings_registry=manager.registry,
            on_exit=AbortAction.RAISE_EXCEPTION,
            on_abort=AbortAction.RETRY,
            ignore_case=True)

        eventloop = create_eventloop()

        self.dcli = CommandLineInterface(
            application=application,
            eventloop=eventloop)

        while True:
            try:
                document = self.dcli.run(True)
                self.handler.handle_input(document.text)

                if isinstance(self.handler.output, GeneratorType):
                    output_stream(self.handler.command,
                                  self.handler.output,
                                  self.handler.log)

                elif self.handler.output is not None:
                    lines = format_data(
                        self.handler.command,
                        self.handler.output)
                    click.echo_via_pager('\n'.join(lines))

                if self.handler.after:
                    for line in self.handler.after():
                        click.echo(line)

                if self.handler.exception:
                    # This was handled, just log it.
                    self.logger.warning('An error was handled: %r',
                                        self.handler.exception)

                self.refresh_completions()

            except OptionError as ex:
                self.logger.debug('Error: %r.', ex)
                self.logger.error("traceback: %r", traceback.format_exc())
                click.secho(ex.msg, fg='red')

            except KeyboardInterrupt:
                # user pressed Ctrl + C
                if self.handler.after:
                    click.echo('')
                    for line in self.handler.after():
                        click.echo(line)

                self.refresh_completions()

            except DockerPermissionException as ex:
                self.logger.debug('Permission exception: %r.', ex)
                self.logger.error("traceback: %r", traceback.format_exc())
                click.secho(ex.message, fg='red')

            except EOFError:
                # exit out of the CLI
                break

            # TODO: uncomment for release
            except Exception as ex:
                self.logger.debug('Exception: %r.', ex)
                self.logger.error("traceback: %r", traceback.format_exc())
                click.secho("{0}".format(ex), fg='red')
                break

        self.revert_less_opts()
        self.write_config_file()
        print('Goodbye!')


@click.command()
@click.option('--no-completion', is_flag=True, default=False, help='Disable autocompletion.')
def cli(no_completion):
    """
    Create and call the CLI
    """
    try:
        dcli = WharfeeCli(no_completion)
        dcli.run_cli()
    except DockerTimeoutException as ex:
        click.secho(ex.message, fg='red')
    except DockerSslException as ex:
        click.secho(ex.message, fg='red')

if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
"""
Helper functions to format output for CLI.
"""
from __future__ import unicode_literals
from __future__ import print_function

import json
import click
import six
from tabulate import tabulate
from pygments import highlight
from pygments.lexers.data import JsonLexer
from pygments.formatters.terminal import TerminalFormatter

# Python 3 has no 'basestring' or 'long' type we're checking for.
try:
    unicode
except NameError:
    # Python 3
    basestring = unicode = str
    long = int


class StreamFormatter(object):

    def __init__(self, data):
        """
        Initialize the formatter passing in the stream.
        :param data: generator
        """
        self.stream = data
        self.counter = 0

    def output(self):
        """
        Process and output line by line.
        :return: int
        """
        for line in self.stream:
            self.counter += 1
            line = line.strip()
            click.echo(line)
        return self.counter


class JsonStreamDumper(StreamFormatter):

    lexer = JsonLexer()
    term = TerminalFormatter()

    def output(self):
        """
        Process and output object by object.
        :return: int
        """
        for obj in self.stream:
            self.counter += 1
            if isinstance(obj, six.string_types):
                click.echo(obj)
            else:
                text = json.dumps(obj, indent=4)
                text = self.colorize(text)
                for line in text.split('\n'):
                    click.echo(line)
        return self.counter

    def colorize(self, text):
        return highlight(text, self.lexer, self.term).rstrip('\r\n')


class JsonStreamFormatter(StreamFormatter):

    progress = False

    def __init__(self, data):
        """
        Initialize the formatter passing in the stream.
        :param data: generator
        """
        StreamFormatter.__init__(self, data)

    def output(self):
        """
        Process and output line by line.
        :return: int

        Lines that contain progress information in them get special handling:

        {
            "status":"Extracting",
            "progressDetail":{
                "current":64618496,
                "total":65771329
            },
            "progress":"[======================\u003e ] 64.62 MB/65.77 MB",
            "id":"e9e06b06e14c"
        }

        """
        for line in self.stream:
            self.counter += 1
            parts = line.strip().decode('utf8').split('\r\n')
            for part in parts:
                data = json.loads(part)
                if self.is_progress(data):
                    self.show_progress_line(data)
                else:
                    self.show_progress_end()
                    self.show_line(data)

        return self.counter

    def is_progress(self, data):
        """
        If the JSON data contains progress bar information.
        :param data: json
        """
        return 'progress' in data and data['progress']

    def show_line(self, data):
        """
        Format and output a JSON line.
        :param data: json
        """
        if 'id' in data and data['id']:
            line = "{0} {1}".format(data['status'], data['id'])
        elif 'status' in data:
            line = "{0}".format(data['status'])
        elif 'stream' in data:
            line = "{0}".format(data['stream'])
        elif 'errorDetail' in data and data['errorDetail']:
            line = "{0}".format(data['errorDetail'].get(
                'message', 'Unknown error'))
        elif 'error' in data and data['error']:
            line = "{0}".format(data['error'])
        else:
            line = "{0}".format(data)

        if line:
            line = line.rstrip()

        click.echo(line)

    def show_progress_end(self):
        """
        If we were just showing progress bar, enter a newline and unset
        the progress flag.
        """
        if self.progress:
            click.echo()
        self.progress = False

    def show_progress_line(self, data):
        """
        Output a carriage return and new progress.
        :param data: json
        """
        click.echo(b'\x0d', nl=False)

        self.progress = True

        line = "{0} {1}: {2}".format(
            data['status'],
            data['id'],
            data['progress'])

        click.echo(line, nl=False)


def format_data(command, data):
    """
    Uses tabulate to format the iterable.
    :return: string (multiline)
    """
    if command and command in DATA_FORMATTERS:
        f = DATA_FORMATTERS[command]
        assert callable(f)
        return f(data)

    if command and command in DATA_FILTERS:
        data = DATA_FILTERS[command](data)

    if isinstance(data, list) and len(data) > 0:
        if isinstance(data[0], tuple):
            if is_plain_lists(data):
                text = tabulate(data)
                return text.split('\n')
            else:
                return format_struct(data)
        elif isinstance(data[0], dict):
            if data[0].keys() == ['Id']:
                # Sometimes our 'quiet' output is a list of dicts but
                # there's only a single "Id" key in each dict. Let's simplify
                # those into plain string lists.
                return [d['Id'] for d in data]
            else:
                data = flatten_rows(data)
                data = truncate_rows(data)
                text = tabulate(data, headers='keys')
                return text.split('\n')
        elif isinstance(data[0], six.string_types):
            if len(data) == 1:
                return data
            elif is_plain_list(data):
                return data
            else:
                data = truncate_rows(data)
                text = tabulate(data)
                return text.split('\n')
    return data


def format_struct(data, spaces=4, indent=0, lines=None):

    if lines is None:
        lines = []

    if isinstance(data, dict):
        data = [(k, data[k]) for k in sorted(data.keys())]

    def item_to_line(current_item, current_line, is_last_item, current_list):
        """ Helper to add item to end of line """
        if len(current_line) == 0:
            current_indent = ' ' * (indent * spaces)
            current_line = current_indent

        current_line += '{0}'.format(current_item)

        if is_last_item:
            current_list.append(current_line)
            current_line = ''
        else:
            current_line += ': '
        return current_line, current_list

    for row in data:
        line = ''
        l = len(row)
        for i in range(l):
            if isinstance(row[i], dict):
                lines.append(line)
                lines = format_struct(row[i], spaces, indent + 1, lines)
            elif isinstance(row[i], list):
                if is_plain_list(row[i]):
                    item = flatten_list(row[i])
                    line, lines = item_to_line(item, line, i == (l - 1), lines)
                else:
                    lines.append(line)
                    lines = format_struct(row[i], spaces, indent + 1, lines)
            else:
                line, lines = item_to_line(row[i], line, i == (l - 1), lines)

    return lines


def is_plain_lists(lst):
    """
    Check if all items in list of lists are strings or numbers
    :param lst:
    :return: boolean
    """
    for x in lst:
        if not is_plain_list(x):
            return False
    return True


def is_plain_list(lst):
    """
    Check if all items in list are strings or numbers
    :param lst:
    :return: boolean
    """
    for item in lst:
        if not isinstance(item, six.string_types) and \
                not isinstance(item, (int, long, float, complex)):
            return False
    return True


def flatten_list(data):
    """
    Format and return a comma-separated string of list items.
    :param data:
    :return:
    """
    return ', '.join(["{0}".format(x) for x in data])


def flatten_dict(data):
    """
    Format and return a comma-separated string of dict items.
    :param data:
    :return:
    """
    return ', '.join(["{0}: {1}".format(x, y) for x, y in data.items()])


def format_port_lines(ports):
    """
    Return ports as list of strings
    :param ports: list of dicts
    :return: list of strings
    """
    port_s = format_ports(ports)
    result = port_s.split(', ')
    return result


def format_ports(ports):
    """
    Ports get special treatment.

    They are a list that looks like this:
       [{u'Type': u'tcp', u'PrivatePort': 3306}]

    We return this:
       "3306/tcp"

    Or instead of this:
       [{
       u'IP': u'0.0.0.0',
       u'Type': u'tcp',
       u'PublicPort': 3000,
       u'PrivatePort': 3306}]

    We return:
        "0.0.0.0:3000->3306/tcp"

    :param ports: list of dicts
    :return: string
    """

    def format_port_list(l):
        if isinstance(l, list):
            return ', '.join([format_port(x) for x in l])
        return '{0}'.format(l)

    def format_port(port):
        """
        Format port dictionary and return string.
        """
        if 'PublicPort' in port:
            return '{0}:{1}->{2}/{3}'.format(
                port.get('IP', '0.0.0.0'),
                port.get('PublicPort', '0000'),
                port.get('PrivatePort', '0000'),
                port.get('Type', 'type'))
        if 'HostPort' in port:
            return "{0}:{1}".format(
                port.get('HostIp', '0.0.0.0'),
                port.get('HostPort', '0000')
            )
        if 'PrivatePort' in port:
            return '{0}/{1}'.format(
                port.get('PrivatePort', '0000'),
                port.get('Type', 'type'))
        # Fallback to formatting "as is"
        return "{0}".format(port)

    if isinstance(ports, dict):
        return ', '.join(['{0}->{1}'.format(k, format_port_list(v))
                          for k, v in ports.items()])
    return ', '.join(format_port(x) for x in ports)


def flatten_rows(rows):
    """
    Transform all list or dict values in a dict into comma-separated strings.
    :param rows: iterable of dictionaries
    :return:
    """

    for row in rows:
        for k in row.keys():
            if k in ROW_FORMATTERS:
                row[k] = ROW_FORMATTERS[k](row[k])
            elif isinstance(row[k], list):
                row[k] = flatten_list(row[k])
            elif isinstance(row[k], dict):
                row[k] = flatten_dict(row[k])
    return rows


def truncate_rows(rows, length=30, length_id=10):
    """
    Truncate every string value in a dictionary up to a certain length.
    :param rows: iterable of dictionaries
    :param length: int
    :param length_id: length for dict keys that end with "Id"
    :return:
    """

    def trimto(s, l):
        """
        Trim string to length.
        :param s: string to trim
        :param l: length
        """
        if isinstance(s, six.string_types):
            return s[:l + 1]
        return s

    result = []
    for row in rows:
        if isinstance(row, dict):
            updated = {}
            for k, v in row.items():
                if k.endswith('Id'):
                    updated[k] = trimto(v, length_id)
                else:
                    updated[k] = trimto(v, length)
            result.append(updated)
        elif isinstance(row, six.string_types):
            result.append(trimto(row, length))
        else:
            result.append(row)
    return result


def format_top(data):
    """
    Format "top" output
    :param data: dict
    :return: list
    """
    result = []
    if data:
        if 'Titles' in data:
            result.append(data['Titles'])
        if 'Processes' in data:
            for process in data['Processes']:
                result.append(process)
    result = tabulate(result, headers='firstrow').split('\n')
    return result


def filter_dict(data, display_keys):
    """
    Strip out some of the dictionary fields.
    :param display_keys: set
    :param data: dict
    :return: dict
    """
    if data and isinstance(data, list) and isinstance(data[0], dict):
        result = []
        for item in data:
            filtered = {}
            for k, v in item.items():
                if k.lower() in display_keys:
                    filtered[k] = v
            result.append(filtered)
        return result
    return data


def filter_ps(data):
    display_keys = set([
        'status', 'created', 'image', 'id', 'command', 'names', 'ports'])
    return filter_dict(data, display_keys)


def filter_volume_ls(data):
    display_keys = set(['driver', 'name'])
    return filter_dict(data, display_keys)


DATA_FILTERS = {
    'ps': filter_ps,
    'volume ls': filter_volume_ls,
}


DATA_FORMATTERS = {
    'top': format_top,
    'port': format_port_lines
}

ROW_FORMATTERS = {
    'Ports': format_ports
}

STREAM_FORMATTERS = {
    'pull': JsonStreamFormatter,
    'push': JsonStreamFormatter,
    'build': JsonStreamFormatter,
    'inspect': JsonStreamDumper,
    'volume inspect': JsonStreamDumper,
}


def output_stream(command, stream, logs):
    """
    Take the iterable and output line by line using click.echo.
    :param command: string command
    :param stream: generator
    :param logs: callable
    :return: None
    """
    if command and command in STREAM_FORMATTERS:
        formatter = STREAM_FORMATTERS[command](stream)
    else:
        formatter = StreamFormatter(stream)

    stream_count = formatter.output()

    if stream_count == 0 and logs and callable(logs):
        # Something nasty has happened and we got an empty
        # output stream. But we have logs. Let's show those.
        lines = logs()
        if lines:
            lines = lines.split('\n')
            for line in lines:
                click.echo(line)



# -*- coding: utf-8
from __future__ import unicode_literals
from __future__ import print_function

from pygments.token import Token


def create_toolbar_handler(is_long_option, is_fuzzy):

    assert callable(is_long_option)
    assert callable(is_fuzzy)

    def get_toolbar_items(_):
        """
        Return bottom menu items
        :param _: cli instance
        :return: list of Token.Toolbar
        """

        if is_long_option():
            option_mode_token = Token.Toolbar.On
            option_mode = 'Long'
        else:
            option_mode_token = Token.Toolbar.Off
            option_mode = 'Short'

        if is_fuzzy():
            fuzzy_token = Token.Toolbar.On
            fuzzy = 'ON'
        else:
            fuzzy_token = Token.Toolbar.Off
            fuzzy = 'OFF'

        return [
            (Token.Toolbar, ' [F2] Help '),
            (option_mode_token, ' [F3] Options: {0} '.format(option_mode)),
            (fuzzy_token, ' [F4] Fuzzy: {0} '.format(fuzzy)),
            (Token.Toolbar, ' [F10] Exit ')
        ]

    return get_toolbar_items



#!/usr/bin/env python
# -*- coding: utf-8
from __future__ import unicode_literals
from __future__ import print_function

import sys
import pretty
import re
import pexpect
import ssl
import six

from docker import AutoVersionClient
from docker.utils import kwargs_from_env
from docker.errors import APIError
from docker.errors import DockerException, InvalidVersion
from requests.exceptions import ConnectionError
from requests.packages.urllib3 import disable_warnings
from .options import allowed_args
from .options import parse_command_options
from .options import format_command_help, format_command_line
from .options import COMMAND_NAMES, split_command_and_args
from .options import OptionError
from .helpers import filesize, parse_port_bindings, parse_volume_bindings, \
    parse_exposed_ports, parse_kv_as_dict
from .utils import shlex_split
from .decorators import if_exception_return


class DockerClient(object):
    """
    This client is a "translator" between docker-py API and
    the standard docker command line. We need one because docker-py
    does not use the same naming for command names and their parameters.
    For example, "docker ps" is named "containers", "-n" parameter
    is named "limit", some parameters are not implemented at all, etc.
    """

    def __init__(self, timeout=None, clear_handler=None, refresh_handler=None, logger=None):
        """
        Initialize the Docker wrapper.
        :param timeout: int
        :param clear_handler: callable
        :param refresh_handler: callable
        :param logger: logger
        """

        assert callable(clear_handler)
        assert callable(refresh_handler)

        self.logger = logger
        self.exception = None

        self.handlers = {
            'attach': (self.attach, 'Attach to a running container.'),
            'build': (self.build, ("Build a new image from the source"
                                   " code")),
            'clear': (clear_handler, "Clear the window."),
            'create': (self.create, 'Create a new container.'),
            'exec': (self.execute, ("Run a command in a running"
                                    " container.")),
            'help': (self.help, "Help on available commands."),
            'pause': (self.pause, "Pause all processes within a container."),
            'ps': (self.containers, "List containers."),
            'port': (self.port, ("List port mappings for the container, or "
                                 "lookup the public-facing port that is "
                                 "NAT-ed to the private_port.")),
            'pull': (self.pull, ("Pull an image or a repository from the "
                                 "registry.")),
            'push': (self.push, ("Push an image or a repository to the "
                                 "registry.")),
            'images': (self.images, "List images."),
            'info': (self.info, "Display system-wide information."),
            'inspect': (self.inspect, "Return low-level information on a " +
                        "container or image."),
            'login': (self.login, ("Register or log in to a Docker registry "
                                   "server (defaut "
                                   "\"https://index.docker.io/v1/\").")),
            'logs': (self.logs, "Fetch the logs of a container."),
            'refresh': (refresh_handler, "Refresh autocompletions."),
            'restart': (self.restart, "Restart a running container."),
            'run': (self.run, "Run a command in a new container."),
            'rm': (self.rm, "Remove one or more containers."),
            'rmi': (self.rmi, "Remove one or more images."),
            'search': (self.search, "Search the Docker Hub for images."),
            'shell': (self.shell, "Get shell into a running container."),
            'start': (self.start, "Restart a stopped container."),
            'stop': (self.stop, "Stop a running container."),
            'tag': (self.tag, "Tag an image into a repository."),
            'top': (self.top, "Display the running processes of a container."),
            'unpause': (self.unpause, ("Unpause all processes within a "
                                       "container.")),
            'version': (self.version, "Show the Docker version information."),
            'volume create': (self.volume_create, "Create a new volume."),
            'volume inspect': (self.volume_inspect, "Inspect one or more "
                               "volumes."),
            'volume ls': (self.volume_ls, "List volumes."),
            'volume rm': (self.volume_rm, "Remove a volume."),
        }

        self.output = None
        self.after = None
        self.command = None
        self.log = None

        self.is_refresh_containers = False
        self.is_refresh_running = False
        self.is_refresh_images = False
        self.is_refresh_volumes = False

        disable_warnings()

        if sys.platform.startswith('darwin') \
                or sys.platform.startswith('win32'):
            try:
                # mac or win
                kwargs = kwargs_from_env()
                # hack from here:
                # http://docker-py.readthedocs.org/en/latest/boot2docker/
                # See also: https://github.com/docker/docker-py/issues/406
                if 'tls' in kwargs:
                    kwargs['tls'].assert_hostname = False
                kwargs['timeout'] = timeout
                self.instance = AutoVersionClient(**kwargs)

            except DockerException as x:
                if 'CERTIFICATE_VERIFY_FAILED' in x.message:
                    raise DockerSslException(x)
                elif 'ConnectTimeoutError' in x.message:
                    raise DockerTimeoutException(x)
                else:
                    raise x
        else:
            # unix-based
            kwargs = kwargs_from_env(
                ssl_version=ssl.PROTOCOL_TLSv1,
                assert_hostname=False)
            kwargs['timeout'] = timeout
            self.instance = AutoVersionClient(**kwargs)

    def debug(self, message):
        """Log a debug message if logger is passed in."""
        if self.logger is not None:
            self.logger.debug(message)

    def handle_input(self, text):
        """
        Parse the command, run it via the client, and return
        some iterable output to print out. This will parse options
        and arguments out of the command line and into parameters
        consistent with docker-py naming. It is designed to be the
        only really public method of the client. Other methods
        are just pass-through methods that delegate commands
        to docker-py.
        :param text: user input
        :return: iterable
        """

        def reset_output():
            """ Set all internals to initial state."""
            self.command = None
            self.is_refresh_containers = False
            self.is_refresh_running = False
            self.is_refresh_images = False
            self.is_refresh_volumes = False
            self.after = None
            self.log = None
            self.exception = None

        tokens = shlex_split(text) if text else ['']
        cmd, params = split_command_and_args(tokens)

        reset_output()

        if cmd and cmd in self.handlers:
            handler = self.handlers[cmd][0]
            self.command = cmd

            if params:
                try:
                    if '-h' in tokens or '--help' in tokens:
                        self.output = [format_command_help(cmd)]
                    else:
                        parser, popts, pargs = parse_command_options(
                            cmd, params)
                        if 'help' in popts:
                            del popts['help']

                        self.output = handler(*pargs, **popts)

                except APIError as ex:
                    reset_output()
                    self.output = [ex.explanation]

                except OptionError as ex:
                    reset_output()
                    raise ex

                except Exception as ex:
                    reset_output()
                    self.output = [ex.__repr__()]
            else:
                self.output = handler()
        elif cmd:
            self.output = self.help()

    def attach(self, *args, **kwargs):
        """
        Attach to a running container.
        :param kwargs:
        :return: None
        """
        if not args:
            return ['Container name or ID is required.']

        container = args[0]

        def on_after():
            self.is_refresh_containers = True
            self.is_refresh_running = True
            return ['\rDetached from {0}.'.format(container)]

        self.after = on_after

        command = format_command_line('attach', False, args, kwargs)
        process = pexpect.spawnu(command)
        process.interact()

    def help(self, *_):
        """
        Collect and return help docstrings for all commands.
        :return: list of tuples
        """

        help_rows = [(key, self.handlers[key][1])
                     for key in COMMAND_NAMES]
        return help_rows

    def not_implemented(self, *_, **kw):
        """
        Placeholder for commands to be implemented.
        :return: iterable
        """
        return ['Not implemented.']

    def version(self, *_):
        """
        Return the version. Equivalent of docker version.
        :return: list of tuples
        """

        try:
            verdict = self.instance.version()
            result = [(k, verdict[k]) for k in sorted(verdict.keys())]
            return result
        except ConnectionError as ex:
            raise DockerPermissionException(ex)

    def info(self, *_):
        """
        Return the system info. Equivalent of docker info.
        :return: list of tuples
        """

        rdict = self.instance.info()
        result = [(k, rdict[k]) for k in sorted(rdict.keys())]
        return result

    def inspect(self, *args, **_):
        """
        Return image or container info. Equivalent of docker inspect.
        :return: dict
        """

        if not args or len(args) == 0:
            yield 'Container or image ID is required.'

        cids, cnames, imids, imnames = set(), set(), set(), set()

        cs = self.containers(all=True)
        if cs and len(cs) > 0 and isinstance(cs[0], dict):
            cids = set([c['Id'] for c in cs])
            cnames = set([name for c in cs for name in c['Names']])

        ims = self.images(all=True)
        if ims and len(ims) > 0 and isinstance(ims[0], dict):
            imids = set([i['Id'] for i in ims])
            imnames = set([i['Repository'] for i in ims])

        for name in args:
            if name in cids or name in cnames:
                info = self.instance.inspect_container(name)
            elif name in imids or name in imnames:
                info = self.instance.inspect_image(name)
            else:
                info = 'Container or image ID is required.'
            yield info

    def containers(self, *_, **kwargs):
        """
        Return the list of containers. Equivalent of docker ps.
        :return: list of dicts
        """

        # Truncate by default.
        if 'trunc' in kwargs and kwargs['trunc'] is None:
            kwargs['trunc'] = True

        def format_names(names):
            """
            Container names start with /.
            Let's strip this for readability.
            """
            if isinstance(names, list):
                formatted = []
                for name in names:
                    if isinstance(name, six.string_types):
                        formatted.append(name.lstrip('/'))
                    else:
                        formatted.append(name)
                return formatted
            return names

        csdict = self.instance.containers(**kwargs)
        if len(csdict) > 0:

            if 'quiet' not in kwargs or not kwargs['quiet']:
                for i in range(len(csdict)):
                    csdict[i]['Names'] = format_names(csdict[i]['Names'])
                    csdict[i]['Created'] = pretty.date(csdict[i]['Created'])

            return csdict
        else:
            return ['There are no containers to list.']

    def pause(self, *args, **kwargs):
        """
        Pause all processes in a container. Equivalent of docker pause.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        kwargs['container'] = args[0]

        self.instance.pause(**kwargs)

        return [kwargs['container']]

    def port(self, *args, **_):
        """
        List port mappings for the container. Equivalent of docker port.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        port_args = [args[0], None]
        port_args[1] = args[1] if len(args) > 1 else None

        result = self.instance.port(*port_args)
        if result:
            return result

        result = self.instance.inspect_container(port_args[0])
        if result:
            result = result.get('NetworkSettings', {}).get('Ports', None)
            if result:
                return result

        return ['There are no port mappings for {0}.'.format(args[0])]

    def rm(self, *args, **kwargs):
        """
        Remove a container. Equivalent of docker rm.
        :param kwargs:
        :return: Container ID or iterable output.
        """

        truncate_output = False

        all_stopped = 'all_stopped' in kwargs and kwargs['all_stopped']
        all = 'all' in kwargs and kwargs['all']

        if all_stopped:
            if args and len(args) > 0:
                return ['Provide either --all-stopped, or container name(s).']

            containers = self.instance.containers(
                quiet=True,
                filters={'status': 'exited'})

            if not containers or len(containers) == 0:
                return ['There are no stopped containers.']

            containers = [c['Id'] for c in containers]
            truncate_output = True

        elif all:
            if args and len(args) > 0:
                return ['Provide either --all, or container name(s).']

            containers = self.instance.containers(quiet=True, all=True)

            if not containers or len(containers) == 0:
                return ['There are no containers.']

            containers = [c['Id'] for c in containers]
            truncate_output = True

        else:
            containers = args

        kwargs = allowed_args('rm', **kwargs)

        def stream():
            for container in containers:
                try:
                    self.instance.remove_container(container, **kwargs)
                    self.is_refresh_containers = True
                    self.is_refresh_running = True
                    if truncate_output:
                        yield "{0:.25}".format(container)
                    else:
                        yield container
                except APIError as ex:
                    yield '{0:.25}: {1}'.format(container, ex.explanation)
                yield 'Removed: {0} container(s).'.format(len(containers) if containers else 0)

        return stream()

    def rmi(self, *args, **kwargs):
        """
        Remove an image. Equivalent of docker rm.
        :param kwargs:
        :return: Image name.
        """

        truncate_output = False

        all_dangling = 'all_dangling' in kwargs and kwargs['all_dangling']
        all = 'all' in kwargs and kwargs['all']

        if all_dangling:
            if args and len(args) > 0:
                return ['Provide either --all-dangling, or image name(s).']

            images = self.instance.images(
                quiet=True,
                filters={'dangling': True})

            if not images or len(images) == 0:
                return ['There are no dangling images.']

            truncate_output = True

        elif all:
            if args and len(args) > 0:
                return ['Provide either --all, or image name(s).']

            images = self.instance.images(quiet=True, all=True)

            if not images or len(images) == 0:
                return ['There are no images.']

            truncate_output = True

        else:
            images = args

        kwargs = allowed_args('rmi', **kwargs)

        def stream():
            for image in images:
                try:
                    self.instance.remove_image(image, **kwargs)
                    self.is_refresh_images = True
                    if truncate_output:
                        yield "{:.25}".format(image)
                    else:
                        yield image
                except APIError as ex:
                    yield '{0:.25}: {1}'.format(image, ex.explanation)

        return stream()

    def run(self, *args, **kwargs):
        """
        Create and start a container. Equivalent of docker run.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Image name is required.']

        if kwargs['remove'] and kwargs['detach']:
            return ['Use either --rm or --detach.']

        # Always call external cli for this, rather than figuring out
        # why docker-py throws "jack is incompatible with use of CloseNotifier in same ServeHTTP call"
        kwargs['force'] = True
        called, args, kwargs = self.call_external_cli('run', *args, **kwargs)
        if not called:
            kwargs['image'] = args[0]
            kwargs['command'] = args[1:] if len(args) > 1 else []

            kwargs = self._add_port_bindings(kwargs)
            kwargs = self._add_exposed_ports(kwargs)
            kwargs = self._add_link_bindings(kwargs)
            kwargs = self._add_volumes_from(kwargs)
            kwargs = self._add_volumes(kwargs)
            kwargs = self._add_network_mode(kwargs)

            create_args = allowed_args('create', **kwargs)
            result = self.instance.create_container(**create_args)

            if result:
                if "Warnings" in result and result['Warnings']:
                    return [result['Warnings']]
                if "Id" in result and result['Id']:
                    self.is_refresh_containers = True
                    is_attach = 'detach' not in kwargs or not kwargs['detach']
                    start_args = allowed_args('start', **kwargs)
                    start_args.update({
                        'container': result['Id'],
                        'attach': is_attach
                    })
                    return self.start(**start_args)
            return ['There was a problem running the container.']

    def create(self, *args, **kwargs):
        """
        Create a container. Equivalent of docker create.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Image name is required.']

        called, args, kwargs = self.call_external_cli('create', *args, **kwargs)
        if not called:
            kwargs['image'] = args[0]
            kwargs['command'] = args[1:] if len(args) > 1 else []

            kwargs = self._add_port_bindings(kwargs)
            kwargs = self._add_exposed_ports(kwargs)
            kwargs = self._add_link_bindings(kwargs)
            kwargs = self._add_volumes_from(kwargs)
            kwargs = self._add_volumes(kwargs)
            kwargs = self._add_network_mode(kwargs)

            create_args = allowed_args('create', **kwargs)
            result = self.instance.create_container(**create_args)

            if result:
                if "Warnings" in result and result['Warnings']:
                    return [result['Warnings']]
                if "Id" in result and result['Id']:
                    self.is_refresh_containers = True
                    return [result['Id']]

            return ['There was a problem creating the container.']

    def restart(self, *args, **kwargs):
        """
        Restart a running container. Equivalent of docker restart.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        def stream():
            for container in args:
                self.instance.restart(container, **kwargs)
                yield container

        return stream()

    @if_exception_return(InvalidVersion, None)
    def volume_create(self, *args, **kwargs):
        """
        Create a volume. Equivalent of docker volume create.
        :param kwargs:
        :return: Volume name.
        """
        if not kwargs:
            return ['Volume name is required.']

        allowed = allowed_args('volume create', **kwargs)

        allowed = self._add_opts(allowed)

        result = self.instance.create_volume(**allowed)
        self.is_refresh_volumes = True
        return [result['Name']]

    @if_exception_return(InvalidVersion, None)
    def volume_ls(self, *args, **kwargs):
        """
        List volumes. Equivalent of docker volume ls.
        :param kwargs:
        :return: Iterable.
        """
        quiet = kwargs.pop('quiet', False)

        kwargs = self._add_filters(kwargs)

        vdict = self.instance.volumes(**kwargs)
        result = vdict.get('Volumes', None)

        if result:
            if quiet:
                result = [volume['Name'] for volume in result]
            return result
        else:
            return ['There are no volumes to list.']

    @if_exception_return(InvalidVersion, None)
    def volume_rm(self, *args, **kwargs):
        """
        Remove a volume. Equivalent of docker volume rm.
        :param kwargs:
        :return: Volume name.
        """
        if not args:
            return ['Volume name is required.']

        def stream():
            for volume in args:
                try:
                    self.instance.remove_volume(volume)
                    self.is_refresh_volumes = True
                    yield volume
                except APIError as x:
                    yield 'Could not remove volume {0}: {1}.'.format(
                        volume,
                        x.explanation)

        return stream()

    @if_exception_return(InvalidVersion, None)
    def volume_inspect(self, *args, **_):
        """
        Return volume info. Equivalent of docker volume ls.
        :return: dict
        """

        if not args or len(args) == 0:
            yield 'Volume name is required.'

        vnames = self.volume_ls(quiet=True)

        for vname in args:
            if vname in vnames:
                info = self.instance.inspect_volume(vname)
                yield info
            else:
                yield "Volume not found: {0}".format(vname)

    def tag(self, *args, **kwargs):
        """
        Tag an image into repository. Equivalent of docker tag.
        :param kwargs:
        :return: Iamge ID.
        """
        if not args or len(args) < 2:
            return ['Image name and repository name are required.']

        img = args[0]
        if ':' in args[1]:
            repo, tag = args[1].rsplit(':', 1)
        else:
            repo, tag = args[1], None

        result = self.instance.tag(
            image=img, repository=repo, tag=tag, **kwargs)

        if result:
            return ['Tagged {0} into {1}.'.format(*args)]
        else:
            return ['Error tagging {0} into {1}.'.format(*args)]

    def _add_filters(self, params):
        """
        Update kwargs if filters are present.
        :param params: dict
        :return dict
        """
        if params.get('filters', None):
            filters = parse_kv_as_dict(params['filters'], True)
            params['filters'] = filters
        return params

    def _add_opts(self, params):
        """
        Update kwargs if opts are present.
        :param params: dict
        :return dict
        """
        if params.get('driver_opts', None):
            opts = parse_kv_as_dict(params['driver_opts'], False)
            params['driver_opts'] = opts
        return params

    def _add_volumes(self, params):
        """
        Update kwargs if volumes are present.
        :param params: dict
        :return dict
        """
        if params.get('volumes', None):
            binds = parse_volume_bindings(params['volumes'])
            params['volumes'] = [x['bind'] for x in binds.values()]
            conf = self.instance.create_host_config(binds=binds)
            self._update_host_config(params, conf)
        return params

    def _add_volumes_from(self, params):
        """
        Update kwargs if volumes-from are present.
        :param params: dict
        :return dict
        """
        if params.get('volumes_from', None):
            cs = ','.join(params['volumes_from'])
            cs = [x.strip() for x in cs.split(',') if x]
            conf = self.instance.create_host_config(volumes_from=cs)
            self._update_host_config(params, conf)
        return params

    def _add_network_mode(self, params):
        """
        Update kwargs if --net is present.
        :param params: dict
        :return: dict
        """
        if 'net' in params:
            conf = self.instance.create_host_config(network_mode=params['net'])
            self._update_host_config(params, conf)
        return params

    def _add_link_bindings(self, params):
        """
        Update kwargs if user wants to link containers.
        :param params: dict
        :return dict
        """
        if params.get('links', None):
            links = {}
            for link in params['links']:
                link_name, link_alias = link.split(':', 1)
                links[link_name] = link_alias
            link_conf = self.instance.create_host_config(links=links)
            self._update_host_config(params, link_conf)
        return params

    def _add_port_bindings(self, params):
        """
        Update kwargs if user wants to bind some ports.
        :param params: dict
        :return dict
        """
        if params.get('port_bindings', None):
            port_bindings = parse_port_bindings(params['port_bindings'])

            # Have to provide list of ports to open in create_container.
            params['ports'] = port_bindings.keys()

            # Have to provide host config with port mappings.
            port_conf = self.instance.create_host_config(
                port_bindings=port_bindings)

            self._update_host_config(params, port_conf)

        return params

    def _add_exposed_ports(self, params):
        """
        Update kwargs if user wants to expose some ports.
        :param params: dict
        :return dict
        """
        if params.get('expose', None):
            ports = parse_exposed_ports(params['expose'])

            # Have to provide list of ports to open in create_container.
            params['ports'] = ports.keys()

            # Have to provide host config with port mappings.
            port_conf = self.instance.create_host_config(
                port_bindings=ports)

            self._update_host_config(params, port_conf)

        return params

    def _update_host_config(self, params, config_to_merge):
        """
        Update config dictionary in kwargs with another dictionary.
        :param params: dict
        :param config_to_merge: dict with new values
        :return dict
        """
        if params.get('host_config', None):
            params['host_config'].update(config_to_merge)
        else:
            params['host_config'] = config_to_merge
        return params

    def _is_repo_tag_valid(self, repo):
        """
        When an image is tagged into a repo, make sure only allowed symbols
        are used.
        :param repo:
        :return: (boolean, "error message")
        """
        # Username: only [a-z0-9_] are allowed, size between 4 and 30
        if '/' not in repo:
            return False, 'Format: user_name/repository_name[:tag].'

        user_name, repo_name = repo.split('/')
        user_pattern = re.compile(r'^[a-z0-9_]{4,30}$')
        if not user_pattern.match(user_name):
            return False, 'Only [a-z0-9_] are allowed in user name, ' \
                          'size between 4 and 30'
        return True, None

    def execute(self, *args, **kwargs):
        """
        Execute a command in the container. Equivalent of docker exec.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args or len(args) < 2:
            return ['Container ID and command is required.']

        called, args, kwargs = self.call_external_cli('exec', *args, **kwargs)
        if not called:
            kwargs['container'] = args[0]
            kwargs['cmd'] = args[1:]

            is_detach = kwargs.pop('detach')

            exec_args = allowed_args('exec', **kwargs)
            result = self.instance.exec_create(**exec_args)

            if result and 'Id' in result:
                return self.instance.exec_start(
                    result['Id'],
                    detach=is_detach,
                    stream=True)

            return ['There was a problem executing the command.']

    def build(self, *args, **kwargs):
        """
        Build an image. Equivalent of docker build.
        :param kwargs:
        :return: Iterable output.
        """
        if not args:
            return ['Directory path or URL is required.']

        kwargs['path'] = args[0]
        kwargs['rm'] = bool(kwargs['rm'])

        self.is_refresh_images = True

        return self.instance.build(**kwargs)

    def shell(self, *args, **_):
        """
        Get the shell into a running container. A shortcut for
        docker exec -it /usr/bin/env bash.
        :param kwargs:
        :return: None
        """
        if not args:
            return ['Container name or ID is required.']

        container = args[0]

        shellcmd = 'bash'
        if len(args) > 1:
            shellcmd = ' '.join(args[1:])

        self.after = lambda: ['\rShell to {0} is closed.'.format(container)]

        command = 'docker exec -it {0} {1}'.format(container, shellcmd)
        process = pexpect.spawnu(command)
        process.interact()

    def start(self, *args, **kwargs):
        """
        Start a container. Equivalent of docker start.
        :param kwargs:
        :return: Container ID or iterable output.
        """

        if args:
            kwargs['container'] = args[0]

        if not kwargs['container']:
            return ['Container name is required.']

        called, args, kwargs = self.call_external_cli('start', *args, **kwargs)
        if not called:

            if 'remove' in kwargs and kwargs['remove']:
                def on_after():
                    container = kwargs['container']
                    try:
                        self.instance.stop(container)
                        self.instance.remove_container(container)
                        yield "Removed container {0:.25} on exit.".format(
                            container)
                    except APIError as ex:
                        yield "{0:.25}: {1}.".format(container, ex.explanation)

                    self.is_refresh_containers = True
                    self.is_refresh_running = True

                self.after = on_after

            startargs = allowed_args('start', **kwargs)

            attached = None

            if 'attach' in kwargs and kwargs['attach']:
                attached = self.view(
                    container=kwargs['container'],
                    stream=True,
                    stdout=True,
                    stderr=False,
                    logs=False)

            result = self.instance.start(**startargs)

            # Just in case the stream generated no output, let's allow for
            # retrieving the logs. They will be our last resort output.
            self.log = lambda: self.instance.logs(kwargs['container'])

            self.is_refresh_running = True
            if result:
                return [result]
            elif attached:
                return attached
            else:
                return [kwargs['container']]

    def view(self, *_, **kwargs):
        """
        Attach to container STDOUT and / or STDERR.
        Docker-py does not allow attaching to STDIN.
        :param kwargs:
        :return: Iterable output
        """
        result = self.instance.attach(**kwargs)
        return result

    def login(self, *args, **kwargs):
        """
        Register or log in to a Docker registry server.
        :param kwargs:
        :return: None
        """
        self.after = lambda: ['\r']

        command = format_command_line('login', False, args, kwargs)
        process = pexpect.spawnu(command)
        process.interact()

    def logs(self, *args, **kwargs):
        """
        Retrieve container logs. Equivalent of docker logs.
        :param kwargs:
        :return: Iterable output
        """
        if not args:
            return ['Container ID/name is required.']

        kwargs['container'] = args[0]

        result = self.instance.logs(**kwargs)
        if isinstance(result, bytes):
            result = result.decode()
        if not kwargs['stream']:
            result = [result]
        return result

    def images(self, *_, **kwargs):
        """
        Return the list of images. Equivalent of docker images.
        :return: list of dicts
        """
        result = self.instance.images(**kwargs)
        re_digits = re.compile('^[0-9]+$', re.UNICODE)

        def convert_image_dict(a):
            """
            Drop some keys and change some values to pretty-print image dict.
            """
            b = {}
            for k, v in a.items():
                if k not in ['RepoTags', 'RepoDigests', 'Labels', 'Size']:
                    b[k] = v
                if k == 'Created' and v and re_digits.search(str(v)):
                    b[k] = pretty.date(v)
                if k == 'VirtualSize':
                    b[k] = filesize(v)

            # If we have more than one repo tag, return as many dicts

            if a.get('RepoTags', None) is None:
                a['RepoTags'] = ['<none>:<none>']

            for rt in a['RepoTags']:
                repo, tag = rt.rsplit(':', 1)
                c = {}
                c.update(b)
                c['Repository'] = repo
                c['Tag'] = tag
                yield c

        if len(result) > 0:
            if isinstance(result[0], dict):
                converted = []
                for x in result:
                    for y in convert_image_dict(x):
                        converted.append(y)
                return converted
            return result
        else:
            return ['There are no images to list.']

    def search(self, *args, **_):
        """
        Return the list of images matching specified term.
        Equivalent of docker search.
        :return: list of dicts
        """

        if not args or len(args) < 1:
            return "Search term is required."

        result = self.instance.search(args[0])

        if len(result) > 0:
            for res in result:
                # Make results  more readable, like official CLI does.
                if 'is_trusted' in res:
                    res['is_trusted'] = '[OK]' if res['is_trusted'] else ''
                if 'is_official' in res:
                    res['is_official'] = '[OK]' if res['is_official'] else ''
            return result
        else:
            return ['No images were found.']

    def stop(self, *args, **kwargs):
        """
        Stop a running container. Equivalent of docker stop.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        def stream():
            for container in args:
                try:
                    self.instance.stop(container, **kwargs)
                    self.is_refresh_running = True
                    yield container
                except APIError as ex:
                    yield '{0:.25}: {1}'.format(container, ex.explanation)

        return stream()

    def top(self, *args, **kwargs):
        """
        Show top processes in a container. Equivalent of docker rm.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        container = args[0]
        result = self.instance.top(container, **kwargs)
        return result

    def pull(self, *args, **kwargs):
        """
        Pull an image by name. Equivalent of docker pull.
        :param kwargs:
        :return: Container ID or iterable output.
        """

        if not args:
            return ['Image name is required.']

        image = args[0]
        kwargs['stream'] = True
        result = self.instance.pull(image, **kwargs)
        self.is_refresh_images = True

        return result

    def push(self, *args, **kwargs):
        """
        Push an image into repository. Equivalent of docker push.
        :param kwargs:
        :return: interactive.
        """
        if not args or len(args) < 1:
            return ['Image name (tagged) is required.']

        tag_valid, tag_message = self._is_repo_tag_valid(args[0])
        if not tag_valid:
            return [tag_message]

        self.after = lambda: ['\r']

        # TODO: this command didn't have to use pexpect.
        # But it was easier to call the official CLI than try and figure out
        # why requests throw this error:
        # File "venv/wharfee/lib/python2.7/site-packages/requests/packages/
        # urllib3/response.py", line 267, in read
        # raise ReadTimeoutError(self._pool, None, 'Read timed out.')
        # requests.packages.urllib3.exceptions.ReadTimeoutError:
        # HTTPSConnectionPool(host='192.168.59.103', port=2376): Read timed out.
        command = format_command_line('push', False, args, kwargs)
        process = pexpect.spawnu(command)
        process.interact()

    def unpause(self, *args, **kwargs):
        """
        Unpause all processes in a container. Equivalent of docker unpause.
        :param kwargs:
        :return: Container ID or iterable output.
        """
        if not args:
            return ['Container name is required.']

        kwargs['container'] = args[0]

        self.instance.unpause(**kwargs)

        return [kwargs['container']]

    def call_external_cli(self, cmd, *args, **kwargs):
        """
        Call the "official" CLI if needed.
        :param args:
        :param kwargs:
        :return:
        """
        called = False

        is_force = kwargs.pop('force', False)
        is_interactive = kwargs.get('interactive', None)
        is_tty = kwargs.get('tty', None)
        is_attach = kwargs.get('attach', None)
        is_attach_bool = is_attach in [True, False]

        def execute_external():
            """
            Call the official cli
            """
            command = format_command_line(cmd, False, args, kwargs)
            process = pexpect.spawnu(command)
            process.interact()

        def on_after_interactive():
            # \r is to make sure when there is some error output,
            # prompt is back to beginning of line
            self.is_refresh_containers = True
            self.is_refresh_running = True
            return ['\rInteractive terminal is closed.']

        def on_after_attach():
            self.is_refresh_containers = True
            self.is_refresh_running = True
            return ['Container exited.\r']

        if is_force or is_interactive or is_tty or (is_attach and not is_attach_bool):
            self.after = on_after_attach if is_attach else on_after_interactive
            called = True
            execute_external()

        return called, args, kwargs


class DockerPermissionException(Exception):

    def __init__(self, inner_exception):
        self.inner_exception = inner_exception
        self.message = """You don't have the necessary permissions to call Docker API.
Try the following:

  # Add a docker group if it does not exist yet.
  sudo groupadd docker

  # Add the connected user "${USER}" to the docker group.
  # Change the user name to match your preferred user.
  sudo gpasswd -a ${USER} docker

  # Restart the Docker daemon.
  # If you are in Ubuntu 14.04, use docker.io instead of docker
  sudo service docker restart

You may need to reboot the machine.
"""


class DockerSslException(Exception):
    """
    Wrapper to handle SSL: CERTIFICATE_VERIFY_FAILED:
    https://github.com/docker/docker-py/issues/465
    """

    def __init__(self, inner_exception):
        self.inner_exception = inner_exception
        self.message = """Your version of requests library has a problem with OpenSSL.
Try the following:

  brew switch openssl 1.0.1j
"""


class DockerTimeoutException(Exception):
    """
    Wrapper to handle ConnectTimeoutError.
    """

    def __init__(self, inner_exception):
        self.inner_exception = inner_exception
        self.message = """The Docker daemon, or boot2docker \
does not seem to be running."""



# -*- coding: utf-8 -*-
import six
import shlex


def shlex_split(text):
    """
    Wrapper for shlex, because it does not seem to handle unicode in 2.6.
    :param text: string
    :return: list
    """
    if six.PY2:
        text = text.encode('utf-8')
    return shlex.split(text)


def shlex_first_token(text):
    """
    Wrapper for shlex, because it does not seem to handle unicode in 2.6.
    :param text: string
    :return: string
    """
    if six.PY2:
        text = text.encode('utf-8')
    lexer = shlex.shlex(text)
    return lexer.get_token()



# -*- coding: utf-8
from __future__ import unicode_literals

import fuzzyfinder

from itertools import chain
from prompt_toolkit.completion import Completer, Completion
from .options import COMMAND_OPTIONS, COMMAND_NAMES, all_options, find_option, \
    split_command_and_args
from .helpers import list_dir, parse_path, complete_path
from .utils import shlex_split, shlex_first_token


class DockerCompleter(Completer):
    """
    Completer for Docker commands and parameters.
    """

    def __init__(self, containers=None, running=None, images=None, tagged=None,
                 volumes=None, long_option_names=True, fuzzy=False):
        """
        Initialize the completer
        :return:
        """
        self.all_completions = set(COMMAND_NAMES)
        self.containers = set(containers) if containers else set()
        self.running = set(running) if running else set()
        self.images = set(images) if images else set()
        self.tagged = set(tagged) if tagged else set()
        self.volumes = set(volumes) if volumes else set()
        self.long_option_mode = long_option_names
        self.fuzzy = fuzzy
        self.enabled = True

    def set_enabled(self, enabled):
        """
        Setter for enabled/disabled property.
        :param enabled: boolean
        """
        self.enabled = enabled

    def set_volumes(self, volumes):
        """
        Setter for list of available volumes.
        :param volumes: list
        """
        self.volumes = set(volumes) if volumes else set()

    def set_containers(self, containers):
        """
        Setter for list of available containers.
        :param containers: list
        """
        self.containers = set(containers) if containers else set()

    def set_running(self, containers):
        """
        Setter for list of running containers.
        :param containers: list
        """
        self.running = set(containers) if containers else set()

    def set_images(self, images):
        """
        Setter for list of available images.
        :param images: list
        """
        self.images = set(images) if images else set()

    def set_tagged(self, images):
        """
        Setter for list of tagged images.
        :param images: list
        """
        self.tagged = set(images) if images else set()

    def set_long_options(self, is_long):
        """
        Setter for long option names.
        :param is_long: boolean
        """
        self.long_option_mode = is_long

    def get_long_options(self):
        """
        Getter for long option names.
        """
        return self.long_option_mode

    def set_fuzzy_match(self, is_fuzzy):
        """
        Setter for fuzzy match option.
        :param is_fuzzy: boolean
        """
        self.fuzzy = is_fuzzy

    def get_fuzzy_match(self):
        """
        Getter for fuzzy match option.
        :return: boolean
        """
        return self.fuzzy

    def get_completions(self, document, _):
        """
        Get completions for the current scope.
        :param document:
        :param _: complete_event
        """
        if not self.enabled:
            return []

        if DockerCompleter.in_quoted_string(document.text):
            return []

        word_before_cursor = document.get_word_before_cursor(WORD=True)
        words = DockerCompleter.get_tokens(document.text)
        command_name = split_command_and_args(words)[0]

        in_command = (len(words) > 1) or \
                     ((not word_before_cursor) and command_name)

        if in_command:
            previous_word = ''
            previous_start = document.find_start_of_previous_word(WORD=True)

            if previous_start == -len(word_before_cursor):
                previous_start = document.find_start_of_previous_word(
                    WORD=True, count=2)

            if previous_start:
                previous_word = document.text_before_cursor[previous_start:]
                previous_word = previous_word.strip().split()[0]

            params = words[1:] if (len(words) > 1) else []
            completions = DockerCompleter.find_command_matches(
                command_name,
                word_before_cursor,
                previous_word,
                params,
                self.containers,
                self.running,
                self.images,
                self.tagged,
                self.volumes,
                self.long_option_mode,
                self.fuzzy)
        else:
            completions = DockerCompleter.find_matches(
                word_before_cursor,
                self.all_completions,
                self.fuzzy)

        return completions

    @staticmethod
    def find_command_matches(command, word='', prev='', params=None,
                             containers=None, running=None, images=None,
                             tagged=None, volumes=None, long_options=True,
                             fuzzy=False):
        """
        Find all matches in context of the given command.
        :param command: string: command keyword (such as "ps", "images")
        :param word: string: word currently being typed
        :param prev: string: previous word
        :param params: list of command parameters
        :param containers: list of containers
        :param running: list of running containers
        :param images: list of images
        :param tagged: list of tagged images
        :param volumes: list of volumes
        :param long_options: boolean
        :param fuzzy: boolean
        :return: iterable
        """

        params = set(params) if params else set([])
        current_opt = find_option(command, prev) if prev else None

        add_directory = False
        add_filepath = False

        if command in COMMAND_OPTIONS:
            opt_suggestions = []
            if current_opt:
                if current_opt.is_type_container():
                    opt_suggestions = containers
                elif current_opt.is_type_running():
                    opt_suggestions = running
                elif current_opt.is_type_image():
                    opt_suggestions = images
                elif current_opt.is_type_tagged():
                    opt_suggestions = tagged
                elif current_opt.is_type_volume():
                    opt_suggestions = volumes
                elif current_opt.is_type_choice():
                    opt_suggestions = current_opt.choices
                elif current_opt.is_type_dirname():
                    add_directory = True
                elif current_opt.is_type_filepath():
                    add_filepath = True

                for m in DockerCompleter.find_collection_matches(
                        word, opt_suggestions, fuzzy):
                    yield m

            if not opt_suggestions:

                def is_unused(o):
                    """
                    Do not offer options that user already set.
                    Unless user may want to set them multiple times.
                    Example: -e VAR1=value1 -e VAR2=value2.
                    """
                    return o.long_name not in params and o.short_name not in params

                def is_current(o):
                    return word in o.names

                def get_opt_name(t):
                    return t.get_name(long_options)

                positionals = []
                possible_options = [x for x in all_options(command) if is_unused(x)
                                    or is_current(x)
                                    or x.is_multiple]
                named_options = sorted([x for x in possible_options if x.name.startswith('-')],
                                       key=get_opt_name)
                positional_options = [x for x in possible_options if not x.name.startswith('-')]

                named_option_map = {}

                for x in named_options:
                    suggestion = x.get_name(long_options)
                    if suggestion:
                        named_option_map[suggestion] = x.display

                for m in DockerCompleter.find_dictionary_matches(
                        word, named_option_map, fuzzy):
                    yield m

                for opt in positional_options:
                    if opt.is_type_container():
                        positionals = chain(positionals, containers)
                    elif opt.is_type_image():
                        positionals = chain(positionals, images)
                    elif opt.is_type_running():
                        positionals = chain(positionals, running)
                    elif opt.is_type_tagged():
                        positionals = chain(positionals, tagged)
                    elif opt.is_type_volume():
                        positionals = chain(positionals, volumes)
                    elif opt.is_type_choice():
                        positionals = chain(positionals, opt.choices)
                    elif opt.is_type_dirname():
                        add_directory = True
                    elif opt.is_type_filepath():
                        add_filepath = True

                # Also return completions for positional options (images,
                # containers, etc.)
                for m in DockerCompleter.find_collection_matches(
                        word, positionals, fuzzy):
                    yield m

        # Special handling for path completion
        if add_directory:
            for m in DockerCompleter.find_directory_matches(word):
                yield m
        if add_filepath:
            for m in DockerCompleter.find_filepath_matches(word):
                yield m

    @staticmethod
    def find_filepath_matches(word):
        """
        Yield matching directory or file names.
        :param word:
        :return: iterable
        """
        base_path, last_path, position = parse_path(word)
        paths = list_dir(word, dirs_only=False)
        for name in sorted(paths):
            suggestion = complete_path(name, last_path)
            if suggestion:
                yield Completion(suggestion, position)

    @staticmethod
    def find_directory_matches(word):
        """
        Yield matching directory names.
        :param word:
        :return: iterable
        """
        base_dir, last_dir, position = parse_path(word)
        dirs = list_dir(word, dirs_only=True)
        for name in sorted(dirs):
            suggestion = complete_path(name, last_dir)
            if suggestion:
                yield Completion(suggestion, position)

    @staticmethod
    def find_dictionary_matches(word, dic, fuzzy):
        """
        Yield all matching names in dict
        :param dic: dict mapping name to display name
        :param word: string user typed
        :param fuzzy: boolean
        :return: iterable
        """

        if fuzzy:
            for suggestion in fuzzyfinder.fuzzyfinder(word, dic.keys()):
                yield Completion(suggestion, -len(word), dic[suggestion])
        else:
            for name in sorted(dic.keys()):
                if name.startswith(word) or not word:
                    yield Completion(name, -len(word), dic[name])

    @staticmethod
    def find_collection_matches(word, lst, fuzzy):
        """
        Yield all matching names in list
        :param lst: collection
        :param word: string user typed
        :param fuzzy: boolean
        :return: iterable
        """

        if fuzzy:
            for suggestion in fuzzyfinder.fuzzyfinder(word, lst):
                yield Completion(suggestion, -len(word))
        else:
            for name in sorted(lst):
                if name.startswith(word) or not word:
                    yield Completion(name, -len(word))

    @staticmethod
    def find_matches(text, collection, fuzzy):
        """
        Find all matches for the current text
        :param text: text before cursor
        :param collection: collection to suggest from
        :param fuzzy: boolean
        :return: iterable
        """
        text = DockerCompleter.last_token(text).lower()

        for suggestion in DockerCompleter.find_collection_matches(
                text, collection, fuzzy):
            yield suggestion

    @staticmethod
    def get_tokens(text):
        """
        Parse out all tokens.
        :param text:
        :return: list
        """
        if text is not None:
            text = text.strip()
            words = DockerCompleter.safe_split(text)
            return words
        return []

    @staticmethod
    def first_token(text):
        """
        Find first word in a sentence
        :param text:
        :return:
        """
        if text is not None:
            text = text.strip()
            if len(text) > 0:
                try:
                    word = shlex_first_token(text)
                    word = word.strip()
                    return word
                except:
                    # no error, just do not complete
                    pass
        return ''

    @staticmethod
    def last_token(text):
        """
        Find last word in a sentence
        :param text:
        :return:
        """
        if text is not None:
            text = text.strip()
            if len(text) > 0:
                word = DockerCompleter.safe_split(text)[-1]
                word = word.strip()
                return word
        return ''

    @staticmethod
    def safe_split(text):
        """
        Shlex can't always split. For example, "\" crashes the completer.
        """
        try:
            words = shlex_split(text)
            return words
        except:
            return text

    @staticmethod
    def in_quoted_string(text):
        """
        Find last word in a sentence
        :param text:
        :return:
        """
        if text is not None:
            text = text.strip()
            if len(text) > 0 and ('"' in text or "'" in text):
                stack = []
                for char in text:
                    if char in ['"', "'"]:
                        if len(stack) > 0 and stack[-1] == char:
                            stack = stack[:-1]
                        else:
                            stack.append(char)
                return len(stack) > 0
        return False



# -*- coding: utf-8
from __future__ import unicode_literals
from __future__ import print_function

import os
import logging


def create_logger(name, log_file, log_level):
    """
    Create and return logger for package.
    :param name: string logger name
    :param log_file: string log file name
    :param log_level: string
    :return: logger
    """
    logger = logging.getLogger(name)

    level_map = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG
    }

    handler = logging.FileHandler(os.path.expanduser(log_file))

    formatter = logging.Formatter(
        '%(asctime)s (%(process)d/%(threadName)s) '
        '%(name)s %(levelname)s - %(message)s')

    handler.setFormatter(formatter)

    root_logger = logging.getLogger('wharfee')
    root_logger.addHandler(handler)
    root_logger.setLevel(level_map[log_level.upper()])

    root_logger.debug('Initializing wharfee logging.')
    root_logger.debug('Log file %r.', log_file)

    return logger



# -*- coding: utf-8
from __future__ import unicode_literals

from optparse import OptionParser, OptionError, OptionGroup
from .option import CommandOption

COMMAND_NAMES = [
    'attach',
    'build',
    'clear',
    'create',
    'exec',
    'help',
    'images',
    'info',
    'inspect',
    'login',
    'logs',
    'ps',
    'pull',
    'pause',
    'port',
    'push',
    'refresh',
    'restart',
    'run',
    'rm',
    'rmi',
    'search',
    'shell',
    'start',
    'stop',
    'tag',
    'top',
    'unpause',
    'version',
    'volume create',
    'volume inspect',
    'volume ls',
    'volume rm',
]


COMMAND_LENGTH = dict((k, len(k.split(' '))) for k in COMMAND_NAMES if ' ' in k)


OPTION_HELP = CommandOption(
    CommandOption.TYPE_BOOLEAN, '-h', '--help',
    action='store_true',
    dest='help',
    help='Display help for this command.')

OPTION_ATTACH_CHOICE = CommandOption(
    CommandOption.TYPE_CHOICE, '-a', '--attach',
    action='append',
    dest='attach',
    help='Attach to STDIN, STDOUT, or STDERR (can use multiple times).',
    nargs='*',
    choices=['stdin', 'stdout', 'stderr'],
    api_match=False)

OPTION_ATTACH_BOOLEAN = CommandOption(
    CommandOption.TYPE_BOOLEAN, '-a', '--attach',
    action='store_true',
    dest='attach',
    help='Attach container\'s STDOUT and STDERR and ' +
         'forward all signals to the process.',
    api_match=False)

OPTION_ENV = CommandOption(
    CommandOption.TYPE_KEYVALUE, '-e', '--env',
    action='append',
    dest='environment',
    help='Set environment variables.',
    nargs='*')

OPTION_EXPOSE = CommandOption(
    CommandOption.TYPE_PORT_RANGE, None, '--expose',
    action='append',
    dest='expose',
    help=('Expose a port or a range of ports (e.g. '
          '--expose=3300-3310) from the container without '
          'publishing it to your host.'),
    nargs='*',
    api_match=False)

OPTION_CONTAINER_HOSTNAME = CommandOption(
    CommandOption.TYPE_STRING, '-h', '--hostname',
    action='store',
    dest='hostname',
    help='Container host name.')

OPTION_CONTAINER_NAME = CommandOption(
    CommandOption.TYPE_CONTAINER, None, '--name',
    action='store',
    dest='name',
    help='Specify volume name.')

OPTION_VOLUME_NAME = CommandOption(
    CommandOption.TYPE_VOLUME, None, '--name',
    action='store',
    dest='name',
    help='Assign a name to the container.')

OPTION_VOLUME_NAME_POS = CommandOption(
    CommandOption.TYPE_VOLUME, 'name',
    help='Volume name.',
    nargs='+')

OPTION_LINK = CommandOption(
    CommandOption.TYPE_CONTAINER, None, '--link',
    action='append',
    dest='links',
    help=('Add link to another container in the form of '
          '<name|id>:alias. To add multiple links: --link '
          'name1:alias1 --link name2:alias2...'),
    nargs='*',
    api_match=False)

OPTION_PUBLISH_ALL = CommandOption(
    CommandOption.TYPE_BOOLEAN, '-P', '--publish-all',
    action='store_true',
    dest='publish_all_ports',
    help=('Publish all exposed ports to the host '
          'interfaces.'),
    api_match=False)

OPTION_PUBLISH = CommandOption(
    CommandOption.TYPE_PORT_BINDING, '-p', '--publish',
    action='append',
    dest='port_bindings',
    help=('Publish a container\'s port to the host. '
          'Format: ip:hostPort:containerPort or '
          'ip::containerPort or hostPort:containerPort or '
          'containerPort. To add multiple ports: --publish '
          '1111:2222 --publish 3333:4444...'),
    nargs='*',
    api_match=False)

OPTION_INTERACTIVE = CommandOption(
    CommandOption.TYPE_BOOLEAN, '-i', '--interactive',
    action='store_true',
    dest='interactive',
    default=False,
    help='Keep STDIN open even if not attached.',
    api_match=False)

OPTION_TTY = CommandOption(
    CommandOption.TYPE_BOOLEAN, '-t', '--tty',
    action='store_true',
    dest='tty',
    default=False,
    help='Allocate a pseudo-TTY.')

OPTION_RM = CommandOption(
    CommandOption.TYPE_BOOLEAN, None, '--rm',
    action='store_true',
    dest='remove',
    help=('Remove the container when it exits. '
          'Can\'t be used with --detach'),
    api_match=False)

OPTION_VOLUME = CommandOption(
    CommandOption.TYPE_FILEPATH, '-v', '--volume',
    action='append',
    dest='volumes',
    help=('Bind mount a volume (e.g., from the host: -v '
          '/host-path:/container-path, from Docker: '
          '-v /container-path).'),
    nargs='*')

OPTION_VOLUMES_FROM = CommandOption(
    CommandOption.TYPE_CONTAINER, None, '--volumes-from',
    action='append',
    dest='volumes_from',
    help=('Mount volumes from the specified containers. Can '
          'be specified multiple times. Alternatively, can '
          'accept a comma-separated string of container '
          'names.'),
    nargs='*',
    api_match=False)

OPTION_NET = CommandOption(
    CommandOption.TYPE_STRING, None, '--net',
    action='store',
    dest='net',
    help='Network mode for the container. Possible values are "bridge", '
         '"none", "container:<name|id>", "host".',
    choices=['bridge', 'none', 'container:', 'host'],
    api_match=False)

OPTION_FILTERS = CommandOption(
    CommandOption.TYPE_STRING, None, '--filter',
    action='append',
    dest='filters',
    nargs='+',
    help='Provide filter values (i.e. "dangling=true").')

OPTION_OPT = CommandOption(
    CommandOption.TYPE_STRING, '-o', '--opt',
    action='append',
    dest='driver_opts',
    nargs='+',
    help='Set driver specific options (e.g. "tardis=blue").')

OPTION_DRIVER = CommandOption(
    CommandOption.TYPE_STRING, '-d', '--driver',
    action='store',
    help='Specify volume driver name (--driver local).')

OPTION_IMAGE = CommandOption(
    CommandOption.TYPE_IMAGE, None, 'image',
    action='store',
    help='Image ID or name to use.')

OPTION_CMD = CommandOption(
    CommandOption.TYPE_COMMAND, None, 'cmd',
    action='store',
    help='Command to run in a container.',
    nargs='?')

OPTION_COMMAND = CommandOption(
    CommandOption.TYPE_COMMAND, None, 'command',
    action='store',
    help='Command to run in a container.',
    nargs='?')

OPTION_STDIN_OPEN = CommandOption(
    CommandOption.TYPE_BOOLEAN, None, 'stdin_open',
    action='store',
    dest='stdin_open',
    default=False)

OPTION_CONTAINER = CommandOption(
    CommandOption.TYPE_CONTAINER, None, 'container',
    action='store',
    help='Container ID or name to use.')

OPTION_CONTAINER_RUNNING = CommandOption(
    CommandOption.TYPE_CONTAINER_RUN, None, 'container',
    action='store',
    help='Container ID or name to use.')

OPTION_HOST_CONFIG = CommandOption(
    CommandOption.TYPE_OBJECT, None, 'host_config',
    action='store',
    dest='host_config')

OPTION_PORTS = CommandOption(
    CommandOption.TYPE_NUMERIC, None, 'ports',
    action='append',
    dest='ports',
    nargs='*')


COMMAND_OPTIONS = {
    'attach': [
        CommandOption(CommandOption.TYPE_BOOLEAN, None, '--no-stdin',
                      action='store_true',
                      dest='no_stdin',
                      help='Do not attach STDIN.',
                      default=False,
                      api_match=False),
        CommandOption(CommandOption.TYPE_CHOICE, None, '--sig-proxy',
                      action='store',
                      dest='sig_proxy',
                      help='Proxy all received signals to the process.',
                      default=True,
                      api_match=False,
                      choices=['true', 'false']),
        OPTION_CONTAINER_RUNNING,
        CommandOption(CommandOption.TYPE_STRING, None, '--detach-keys',
                      action='store',
                      dest='detach_keys',
                      help='Override the key sequence for detaching a container.',
                      api_match=False),
    ],
    'build': [
        CommandOption(CommandOption.TYPE_IMAGE, '-t', '--tag',
                      dest='tag',
                      help=('Repository name (and optionally a tag) to be '
                            'applied to the resulting image in case of '
                            'success.')),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-q', '--quiet',
                      action='store_true',
                      dest='quiet',
                      help=('Suppress the verbose output generated by the '
                            'containers.')),
        CommandOption(CommandOption.TYPE_CHOICE, '--rm',
                      action='store',
                      dest='rm',
                      help=('Remove intermediate containers after a '
                            'successful build.'),
                      default='true',
                      choices=['true', 'false']),
        CommandOption(CommandOption.TYPE_BOOLEAN, '--no-cache',
                      action='store_true',
                      dest='nocache',
                      help='Do not use cache when building the image.'),
        CommandOption(CommandOption.TYPE_DIRPATH, 'path',
                      help='Path or URL where the Dockerfile is located.'),
    ],
    'clear': [],
    'create': [
        OPTION_ATTACH_CHOICE,
        OPTION_ENV,
        OPTION_EXPOSE,
        OPTION_INTERACTIVE,
        OPTION_LINK,
        OPTION_CONTAINER_HOSTNAME,
        OPTION_CONTAINER_NAME,
        OPTION_PUBLISH_ALL,
        OPTION_PUBLISH,
        OPTION_TTY,
        OPTION_VOLUME,
        OPTION_VOLUMES_FROM,
        OPTION_IMAGE,
        OPTION_COMMAND,
        OPTION_NET,
    ],
    'exec': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-d', '--detach',
                      action='store_true',
                      dest='detach',
                      help='Detached mode: run command in the background.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-i', '--interactive',
                      action='store_true',
                      dest='interactive',
                      default=False,
                      help='Keep STDIN open even if not attached.',
                      api_match=False),
        OPTION_TTY,
        OPTION_CONTAINER_RUNNING,
        OPTION_CMD,
    ],
    'info': [
    ],
    'inspect': [
        CommandOption(CommandOption.TYPE_IMAGE, 'image',
                      action='store',
                      dest="image_id",
                      help='Image to inspect.',
                      nargs='*'),
        CommandOption(CommandOption.TYPE_CONTAINER, 'container',
                      action='store',
                      help='Container to inspect.',
                      nargs='*'),
    ],
    'login': [
        CommandOption(CommandOption.TYPE_STRING, '-e', '--email',
                      help='Email.'),
        CommandOption(CommandOption.TYPE_STRING, '-p', '--password',
                      help='Password.'),
        CommandOption(CommandOption.TYPE_STRING, '-u', '--username',
                      help='Username.'),
        CommandOption(CommandOption.TYPE_STRING, None, 'server',
                      dest='registry',
                      help='Registry server.'),
    ],
    'logs': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-f', '--follow',
                      action='store_true',
                      dest='stream',
                      help='Follow log output.'),
        OPTION_CONTAINER,
    ],
    'pause': [
        OPTION_CONTAINER_RUNNING,
    ],
    'port': [
        OPTION_CONTAINER,
        CommandOption(CommandOption.TYPE_STRING, None, 'port',
                      action='store',
                      dest='port',
                      help='Port number (optionally with protocol).'),
    ],
    'ps': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-a', '--all',
                      action='store_true',
                      dest='all',
                      help='Show all containers. '
                           'Only running containers are shown by default.'),
        CommandOption(CommandOption.TYPE_CONTAINER, None, '--before',
                      action='store',
                      dest='before',
                      help='Show only container created before Id or Name, ' +
                           'include non-running ones.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-l', '--latest',
                      action='store_true',
                      dest='latest',
                      help='Show only the latest created container, ' +
                           'include non-running ones.'),
        CommandOption(CommandOption.TYPE_NUMERIC, '-n', None,
                      action='store',
                      dest='limit',
                      help='Show n last created containers, include ' +
                           'non-running ones.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, None, '--no-trunc',
                      action='store_false',
                      dest='trunc',
                      help='Don\'t truncate output.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-q', '--quiet',
                      action='store_true',
                      dest='quiet',
                      help='Only display numeric IDs.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-s', '--size',
                      action='store_true',
                      dest='latest',
                      help='Display total file sizes.'),
        CommandOption(CommandOption.TYPE_CONTAINER, None, '--since',
                      action='store',
                      dest='since',
                      help='Show only containers created since Id or Name, ' +
                           'include non-running ones.')
    ],
    'pull': [
        CommandOption(CommandOption.TYPE_IMAGE, 'image',
                      action='store',
                      help='Image name to pull.'),
    ],
    'push': [
        CommandOption(CommandOption.TYPE_IMAGE_TAGGED, 'name',
                      action='store',
                      help='Image name to push (format: "name[:tag]").'),
    ],
    'images': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-a', '--all',
                      action='store_true',
                      dest='all',
                      help='Show all images (by default filter out the ' +
                           'intermediate image layers).'),
        CommandOption(CommandOption.TYPE_IMAGE, '-f', '--filter',
                      action='store',
                      dest='name',
                      help='Provide name to filter on.'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-q', '--quiet',
                      action='store_true',
                      dest='quiet',
                      help='Only show numeric IDs.')
    ],
    'refresh': [],
    'run': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-d', '--detach',
                      action='store_true',
                      dest='detach',
                      help=('Detached mode: run the container in the '
                            'background and print the new container ID')),
        OPTION_ATTACH_CHOICE,
        OPTION_ENV,
        OPTION_EXPOSE,
        OPTION_CONTAINER_HOSTNAME,
        OPTION_CONTAINER_NAME,
        OPTION_LINK,
        OPTION_PUBLISH_ALL,
        OPTION_PUBLISH,
        OPTION_INTERACTIVE,
        OPTION_TTY,
        OPTION_RM,
        OPTION_VOLUME,
        OPTION_VOLUMES_FROM,
        CommandOption(CommandOption.TYPE_IMAGE_TAGGED, None, 'image',
                      action='store',
                      help='Image name.'),
        OPTION_COMMAND,
        OPTION_NET,
    ],
    'shell': [
        OPTION_CONTAINER_RUNNING,
        CommandOption(CommandOption.TYPE_CHOICE, 'command',
                      action='store',
                      help='Shell command to execute.',
                      choices=[
                          'bash',
                          'sh',
                          'zsh',
                          '/bin/sh',
                          '/usr/bin/bash',
                          '/usr/bin/sh',
                          '/usr/bin/zsh',
                          '/usr/local/bin/bash',
                          '/usr/local/bin/sh',
                          '/usr/local/bin/zsh',
                      ],
                      default='bash',
                      nargs='?'),
    ],
    'start': [
        OPTION_ATTACH_BOOLEAN,
        CommandOption(CommandOption.TYPE_BOOLEAN, '-i', '--interactive',
                      action='store_true',
                      dest='interactive',
                      default=False,
                      help='Attach container\'s STDIN.',
                      api_match=False),
        OPTION_CONTAINER,
    ],
    'restart': [
        CommandOption(CommandOption.TYPE_NUMERIC, '-t', '--timeout',
                      dest='timeout',
                      help=('Number of seconds to try to stop for before '
                            'killing the container. Once killed it will then '
                            'be restarted. Default is 10 seconds.')),
        CommandOption(CommandOption.TYPE_CONTAINER, 'container',
                      action='store',
                      help='Container ID or name to use.',
                      nargs='+'),
    ],
    'rm': [
        CommandOption(CommandOption.TYPE_CONTAINER, 'container',
                      action='store',
                      help='Container ID or name to use.',
                      nargs='+'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '--all-stopped',
                      action='store_true',
                      dest='all_stopped',
                      help='Shortcut to remove all stopped containers.',
                      api_match=False,
                      cli_match=False),
        CommandOption(CommandOption.TYPE_BOOLEAN, '--all',
                      action='store_true',
                      dest='all',
                      help='Shortcut to remove all containers.',
                      api_match=False,
                      cli_match=False),
        CommandOption(CommandOption.TYPE_BOOLEAN, '-f', '--force',
                      action='store_true',
                      dest='force',
                      help='Force the removal of a running container (uses SIGKILL).'),
    ],
    'rmi': [
        CommandOption(CommandOption.TYPE_IMAGE_TAGGED, 'image',
                      action='store',
                      help='Image name name to remove.',
                      nargs='+'),
        CommandOption(CommandOption.TYPE_BOOLEAN, '--all-dangling',
                      action='store_true',
                      dest='all_dangling',
                      help='Shortcut to remove all dangling images.',
                      api_match=False,
                      cli_match=False),
        CommandOption(CommandOption.TYPE_BOOLEAN, '--all',
                      action='store_true',
                      dest='all',
                      help='Shortcut to remove all images.',
                      api_match=False,
                      cli_match=False),
    ],
    'search': [
        CommandOption(CommandOption.TYPE_IMAGE, 'term',
                      action='store',
                      help='A term to search for.'),
    ],
    'stop': [
        OPTION_CONTAINER_RUNNING,
    ],
    'tag': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-f', '--force',
                      action='store_true',
                      dest='force',
                      help='Force.'),
        CommandOption(CommandOption.TYPE_IMAGE, 'image',
                      action='store',
                      help='The image to tag (format: "image[:tag]").'),
        CommandOption(CommandOption.TYPE_IMAGE_TAG, 'name',
                      action='store',
                      help=('The tag name (format: "[registryhost/]'
                            '[username/]name[:tag]").')),
    ],
    'top': [
        OPTION_CONTAINER_RUNNING,
    ],
    'unpause': [
        OPTION_CONTAINER_RUNNING,
    ],
    'volume create': [
        OPTION_VOLUME_NAME,
        OPTION_DRIVER,
        OPTION_OPT
    ],
    'volume inspect': [
        OPTION_VOLUME_NAME_POS
    ],
    'volume ls': [
        CommandOption(CommandOption.TYPE_BOOLEAN, '-q', '--quiet',
                      action='store_true',
                      dest='quiet',
                      help='Only display volume names.'),
        OPTION_FILTERS
    ],
    'volume rm': [
        OPTION_VOLUME_NAME_POS
    ],
}


# Hidden options are options that docker-py supports, but the standard docker cli doesn't have
# them. Since we're emulating the standard cli as close as possible, we're not suggesting these
# to the user.
HIDDEN_OPTIONS = {
    'start': [
        OPTION_PUBLISH_ALL,
    ],
    'run': [
        OPTION_PORTS,
        OPTION_HOST_CONFIG,
        OPTION_STDIN_OPEN
    ],
    'create': [
        OPTION_PORTS,
        OPTION_HOST_CONFIG,
        OPTION_STDIN_OPEN
    ]
}


def all_option_names():
    """
    Helper method to go through all commands and return all option names,
    long or short.
    :return: iterable
    """
    opts = set([OPTION_HELP.short_name, OPTION_HELP.long_name])
    for command in COMMAND_OPTIONS:
        for opt in COMMAND_OPTIONS[command]:
            if opt.short_name and opt.short_name.startswith('-'):
                opts.add(opt.short_name)
            if opt.long_name and opt.long_name.startswith('--'):
                opts.add(opt.long_name)
    return sorted(list(opts))


def find_option(command, name):
    """
    Helper method to find command option by its name.
    :param command: string
    :param name: string
    :return: CommandOption
    """
    # TODO: use all_options
    if command in COMMAND_OPTIONS:
        if name == 'help':
            return OPTION_HELP
        for opt in COMMAND_OPTIONS[command]:
            if name in [opt.short_name, opt.long_name]:
                return opt
    return None


def allowed_args(command_name, **kwargs):
    """
    Return only arguments that the command accepts.
    :param command_name: string
    :param kwargs: dict
    :return: dict
    """
    matches = {}
    available = all_supported(command_name)
    if available:
        for k in kwargs:
            if k in available:
                matches[k] = kwargs[k]
    return matches


def all_options(command, include_hidden=False):
    """
    Helper method to find all command options.
    :param command: string
    :param include_hidden: boolean
    :return: set of CommandOption
    """
    result = [OPTION_HELP]
    if command in COMMAND_OPTIONS:
        result.extend(COMMAND_OPTIONS[command])
    if include_hidden and command in HIDDEN_OPTIONS:
        result.extend(HIDDEN_OPTIONS[command])
    return result


def all_supported(command):
    """
    Helper method to find all command options that docker-py supports.
    :param command: string
    :return: set of CommandOption
    """
    result = set([OPTION_HELP])

    if command in COMMAND_OPTIONS:
        result.update(
            [x.dest for x in COMMAND_OPTIONS[command] if x.api_match])

    if command in HIDDEN_OPTIONS:
        result.update([x.dest for x in HIDDEN_OPTIONS[command]])

    return result


def parse_command_options(cmd, params):
    """
    Parse options for a given command.
    :param cmd: string: command name
    :param params: list: all tokens after command name
    :return: parser, args, opts
    """
    parser = OptParser(
        prog=cmd, add_help_option=False, conflict_handler='resolve')
    parser.disable_interspersed_args()
    for opt in all_options(cmd, include_hidden=True):
        if opt.name.startswith('-'):
            parser.add_option(*opt.args, **opt.kwargs)
    popts, pargs = parser.parse_args(params)
    parser.assert_option_format()
    popts = vars(popts)

    # Add hidden defaults
    if cmd in HIDDEN_OPTIONS:
        for opt in HIDDEN_OPTIONS[cmd]:
            if opt.default is not None:
                popts[opt.dest] = opt.default

    return parser, popts, pargs


def format_command_line(cmd, is_long, args, kwargs):
    """
    Reconstruct the command  line for sending back to official Docker CLI.
    :param cmd: command
    :param is_long: bool use long option names
    :param args: positional parameters
    :param kwargs: named parameters
    :return: string
    """
    opts = dict([(x.dest, x) for x in all_options(cmd)
                 if x.name.startswith('-')])
    comps = ['docker', cmd]

    def kve(o, v):
        v1, v2 = v.split('=', 1)
        if ' ' in v2:
            return '{0} {1}="{2}"'.format(o.get_name(is_long), v1, v2)
        return '{0} {1}={2}'.format(o.get_name(is_long), v1, v2)

    def kv(o, v):
        if o.dest == 'environment':
            return kve(o, v)
        return '{0}={1}'.format(o.get_name(is_long), v)

    for opt_dest, opt_value in kwargs.items():
        if opt_dest in opts and opt_value is not None:
            opt = opts[opt_dest]
            if opt.cli_match:
                if isinstance(opt_value, bool):
                    # skip appending if default
                    if opt.default is not None and opt_value == opt.default:
                        continue
                    comps.append(opt.get_name(is_long))
                elif isinstance(opt_value, list):
                    comps.append(' '.join([kv(opt, val) for val in opt_value]))
                else:
                    comps.append(kv(opt, opt_value))
    comps.extend(args)
    external_command = ' '.join(comps)
    return external_command


def split_command_and_args(tokens):
    """
    Take all tokens from command line, return command part and args part.
    Command can be more than 1 words.
    :param tokens: list
    :return: tuple of (string, list)
    """
    command, args = None, None
    if tokens:
        length = 1
        for cmd_name in COMMAND_LENGTH:
            if ' '.join(tokens).startswith(cmd_name):
                length = COMMAND_LENGTH[cmd_name]
                break
        command = ' '.join(tokens[:length])
        args = tokens[length:] if len(tokens) >= length else None
    return command, args


def format_command_help(cmd):
    """
    Format help string for the command.
    :param cmd: string: command name
    :return: string
    """
    usage = [cmd, '[options]']
    alls = all_options(cmd)

    standards = [_ for _ in alls if _.cli_match]
    extras = [_ for _ in alls if not _.cli_match]

    for opt in alls:
        if not opt.name.startswith('-'):
            optname = "[{0}]".format(opt.name) if opt.is_optional else opt.name
            usage.append(optname)

    usage = ' '.join(usage)

    parser = OptParser(prog=cmd, add_help_option=False, usage=usage,
                       conflict_handler='resolve')

    for opt in standards:
        if opt.name.startswith('-'):
            parser.add_option(*opt.args, **opt.kwargs)

    if extras:
        g = OptionGroup(parser, "Non-standard options")
        for opt in extras:
            g.add_option(*opt.args, **opt.kwargs)
        parser.add_option_group(g)

    return parser.format_help()


class OptParser(OptionParser):

    # TODO: Bad bad bad. There should be a better way to do this.
    def assert_option_format(self):
        """
        I don't want environment vars to be provided as
        "-e KEY VALUE", I want "-e KEY=VALUE" instead.
        Would argparse help here?
        """
        dict_values = vars(self.values)
        if 'environment' in dict_values and dict_values['environment']:
            for envar in dict_values['environment']:
                if '=' not in envar:
                    raise OptionError(
                        'Usage: -e KEY1=VALUE1 -e KEY2=VALUE2...',
                        '-e')

    """
    Wrapper around OptionParser.
    Overrides error method to throw an exception.
    """
    def error(self, msg):
        """error(msg : string)

        Print a usage message incorporating 'msg' to stderr and exit.
        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        raise Exception("Error parsing options: {0}".format(msg))



# -*- coding: utf-8
import shutil
from os.path import expanduser, exists

from configobj import ConfigObj


def read_config(usr_config, def_config=None):
    """
    Read config file (if not exists, read default config).
    :param usr_config: string: config file name
    :param def_config: string: default name
    :return: ConfigParser
    """
    usr_config_file = expanduser(usr_config)

    cfg = ConfigObj()
    cfg.filename = usr_config_file

    if def_config:
        cfg.merge(ConfigObj(def_config, interpolation=False))

    cfg.merge(ConfigObj(usr_config_file, interpolation=False))
    return cfg


def write_default_config(source, destination, overwrite=False):
    """
    Write default config (from template).
    :param source: string: path to template
    :param destination: string: path to write
    :param overwrite: boolean
    """
    destination = expanduser(destination)
    if not overwrite and exists(destination):
        return

    shutil.copyfile(source, destination)



# -*- coding: utf-8
from pygments.token import Token
from pygments.util import ClassNotFound
from prompt_toolkit.styles import default_style_extensions, PygmentsStyle
import pygments.styles


def style_factory(name):
    try:
        style = pygments.styles.get_style_by_name(name)
    except ClassNotFound:
        style = pygments.styles.get_style_by_name('native')

    styles = {}

    styles.update(style.styles)
    styles.update(default_style_extensions)
    styles.update({
        Token.Menu.Completions.Completion.Current: 'bg:#00aaaa #000000',
        Token.Menu.Completions.Completion: 'bg:#008888 #ffffff',
        Token.Menu.Completions.ProgressButton: 'bg:#003333',
        Token.Menu.Completions.ProgressBar: 'bg:#00aaaa',
        Token.Toolbar: 'bg:#222222 #cccccc',
        Token.Toolbar.Off: 'bg:#222222 #004444',
        Token.Toolbar.On: 'bg:#222222 #ffffff',
        Token.Toolbar.Search: 'noinherit bold',
        Token.Toolbar.Search.Text: 'nobold',
        Token.Toolbar.System: 'noinherit bold',
        Token.Toolbar.Arg: 'noinherit bold',
        Token.Toolbar.Arg.Text: 'nobold'
    })

    return PygmentsStyle.from_defaults(style_dict=styles)



# -*- coding: utf-8
import functools


def if_exception_return(ex_type, then_result):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            try:
                return f(self, *args, **kwargs)
            except ex_type as x:
                self.exception = x
                return then_result
        return wrapper
    return decorator



# -*- coding: utf-8
from __future__ import unicode_literals


class CommandOption(object):
    """
    Wrapper for the optparse's option that adds some extra fields
    useful to do autocompletion in context of this option.
    """

    choices = None

    OPTION_VALUES = range(18)

    TYPE_FILEPATH, \
        TYPE_DIRPATH, \
        TYPE_BOOLEAN, \
        TYPE_NUMERIC, \
        TYPE_CONTAINER, \
        TYPE_CONTAINER_RUN, \
        TYPE_IMAGE, \
        TYPE_IMAGE_TAGGED, \
        TYPE_IMAGE_TAG, \
        TYPE_COMMAND, \
        TYPE_COMMAND_ARG, \
        TYPE_CHOICE, \
        TYPE_KEYVALUE, \
        TYPE_PORT_BINDING, \
        TYPE_PORT_RANGE, \
        TYPE_OBJECT, \
        TYPE_STRING,\
        TYPE_VOLUME = \
        OPTION_VALUES

    def __init__(self, option_type, short_name, long_name=None, **kwargs):
        """
        Constructor for the CommandOption
        :param option_type: int: one the type constants from above
        :param short_name: string: short name to create optparse's Option
        :param long_name: string: long name to create optparse's Option
        :param kwargs: keyword args
        :return:
        """
        if option_type not in self.OPTION_VALUES:
            raise ValueError("Incorrect option_type.", option_type)

        if long_name and short_name:
            arguments = [short_name, long_name]
            self.display = '{0}/{1}'.format(short_name, long_name)
        elif long_name:
            arguments = [long_name]
            self.display = long_name
        else:
            arguments = [short_name]
            self.display = short_name

        if 'api_match' in kwargs and kwargs['api_match'] is not None:
            self.api_match = kwargs['api_match']
            del kwargs['api_match']
        else:
            self.api_match = True

        if 'cli_match' in kwargs and kwargs['cli_match'] is not None:
            self.cli_match = kwargs['cli_match']
            del kwargs['cli_match']
        else:
            self.cli_match = True

        if 'dest' in kwargs:
            self.dest = kwargs['dest']
        elif long_name:
            self.dest = long_name.strip('-')
        else:
            self.dest = short_name.strip('-')

        self.default = kwargs['default'] if 'default' in kwargs else None

        if 'nargs' in kwargs:
            self.is_optional = (kwargs['nargs'] in ['?', '*'])
            self.is_multiple = (kwargs['nargs'] in ['+', '*'])

            # TODO: Optparse wants a number here... back to argparse?
            if kwargs['nargs'] in ['?', '*', '+']:
                if kwargs['nargs'] in ['*', '+']:
                    kwargs['action'] = 'append'
                del kwargs['nargs']

        else:
            self.is_optional = False
            self.is_multiple = False

        if 'choices' in kwargs:
            self.choices = kwargs['choices']
            if option_type != CommandOption.TYPE_CHOICE:
                kwargs.pop('choices')

        self.option_type = option_type
        self.short_name = short_name
        self.long_name = long_name
        self.args = arguments
        self.kwargs = kwargs

    def is_type_choice(self):
        """
        If this option is a list of choices.
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_CHOICE or self.choices

    def is_type_container(self):
        """
        Should this option suggest container name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_CONTAINER

    def is_type_running(self):
        """
        Should this option suggest running container name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_CONTAINER_RUN

    def is_type_image(self):
        """
        Should this option suggest image name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_IMAGE

    def is_type_tagged(self):
        """
        Should this option suggest tagged image name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_IMAGE_TAGGED

    def is_type_volume(self):
        """
        Should this option suggest volume name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_VOLUME

    def is_type_filepath(self):
        """
        Should this option suggest filename?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_FILEPATH

    def is_type_dirname(self):
        """
        Should this option suggest directory name?
        :return: boolean
        """
        return self.option_type == CommandOption.TYPE_DIRPATH

    def get_name(self, is_long):
        """
        Return short name if we have one, and is requested. Otherwise default
        to long name.
        :param is_long: boolean
        :return: string
        """
        if is_long:
            return self.long_name if self.long_name else self.short_name
        return self.short_name if self.short_name else self.long_name

    def is_match(self, word):
        """
        Can this option be suggested having this word being typed?
        :param word:
        :return:
        """
        if word:
            return (self.long_name and self.long_name.startswith(word)) or \
                   (self.short_name and self.short_name.startswith(word))
        return True

    @property
    def name(self):
        """
        Getter for short name
        :return: string
        """
        return self.long_name if self.long_name else self.short_name

    @property
    def names(self):
        """
        Getter for all possible names.
        :return: list
        """
        if self.short_name and self.long_name:
            return [self.short_name, self.long_name]
        elif self.long_name:
            return [self.long_name]
        else:
            return [self.short_name]

    def __repr__(self):
        """
        Return the printable representation.
        """
        return 'CommandOption({0}, {1}'.format(
            self.short_name,
            self.long_name)



# -*- coding: utf-8
import os
import math


def parse_kv_as_dict(filters, convert_boolean=False):
    """
    Parse list of "key=value" into dict
    :param filters: list
    :param convert_boolean: boolean
    :return: dict
    """
    result = {}
    if filters:
        for x in filters:
            k, v = x.split('=', 2)
            if convert_boolean:
                if v.lower() == 'true':
                    v = True
                elif v.lower() == 'false':
                    v = False
            result[k] = v
    return result


def parse_volume_bindings(volumes):
    """
    Parse volumes into a dict.
    :param volumes: list of strings
    :return: dict
    """

    def parse_volume(v):
        if ':' in v:
            parts = v.split(':')
            if len(parts) > 2:
                hp, cp, ro = parts[0], parts[1], parts[2]
                return hp, cp, ro == 'ro'
            else:
                hp, cp = parts[0], parts[1]
                return hp, cp, False
        else:
            return None, v, False

    result = {}
    if volumes:
        for vol in volumes:
            host_path, container_path, read_only = parse_volume(vol)
            if host_path:
                result[host_path] = {
                    'bind': container_path,
                    'ro': read_only
                }
    return result


def parse_exposed_ports(ps):
    """
    Parse array of exposed ports (not public).

    [1000] -> { 1000: None }
    [1000-1002] -> { 1000: None, 1001: None, 1002: None }

    :return: dict
    """
    result = {}
    for p in ps:
        if '-' in p:
            # it is a range from port to port
            p1, p2 = p.split('-')
            for x in range(p1, p2 + 1):
                result[x] = None
        else:
            result[p] = None
    return result


def parse_port_bindings(bindings):
    """
    Parse array of string port bindings into a dict. For example:
        ['4567:1111', '2222']
    becomes
        port_bindings={
            1111: 4567,
            2222: None
        }
    and
        ['127.0.0.1:4567:1111']
    becomes
        port_bindings={
            1111: ('127.0.0.1', 4567)
        }
    :param bindings: array of string
    :return: dict
    """

    def parse_port_mapping(s):
        """
        Parse single port mapping.
        """
        if ':' in s:
            parts = s.split(':')
            if len(parts) > 2:
                ip, hp, cp = parts[0], parts[1], parts[2]
                return cp, (ip, None if hp == '' else hp)
            else:
                hp, cp = parts[0], parts[1]
                return cp, None if hp == '' else hp
        else:
            return s, None

    result = {}
    if bindings:
        for binding in bindings:
            container_port, mapping = parse_port_mapping(binding)
            result[container_port] = mapping
    return result


def filesize(size):
    """
    Pretty-print file size from bytes.
    """
    size_name = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    if int(size) > 0:
        i = int(math.floor(math.log(size, 1024)))
        p = math.pow(1024, i)
        s = round(size / p, 3)
        if s > 0:
            return '%s %s' % (s, size_name[i])
    return '0 B'


def complete_path(curr_dir, last_dir):
    """
    Return the path to complete that matches the last entered component.
    If the last entered component is ~, expanded path would not
    match, so return all of the available paths.
    :param curr_dir: string
    :param last_dir: string
    :return: string
    """
    if not last_dir or curr_dir.startswith(last_dir):
        return curr_dir
    elif last_dir == '~':
        return os.path.join(last_dir, curr_dir)


def parse_path(root_dir):
    """
    Split path into head and last component for the completer.
    Also return position where last component starts.
    :param root_dir: path
    :return: tuple of (string, string, int)
    """
    base_dir, last_dir, position = '', '', 0
    if root_dir:
        base_dir, last_dir = os.path.split(root_dir)
        position = -len(last_dir) if last_dir else 0
    return base_dir, last_dir, position


def list_dir(root_dir, dirs_only=False, include_special=False):
    """
    List directory.
    :param root_dir: string: directory to list
    :param dirs_only: boolean
    :param include_special: boolean
    :return: list
    """
    root_dir = '.' if not root_dir else root_dir
    res = []

    if '~' in root_dir:
        root_dir = os.path.expanduser(root_dir)

    if not os.path.exists(root_dir):
        root_dir, _ = os.path.split(root_dir)

    if os.path.exists(root_dir):
        for name in os.listdir(root_dir):
            path = os.path.join(root_dir, name)
            if not include_special and name.startswith('.'):
                continue
            if dirs_only and not os.path.isdir(path):
                continue
            res.append(name)
    return res



from wharfee.__init__ import __version__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    description='Wharfee: a shell for Docker',
    author='Iryna Cherniavska',
    url='http://wharfee.com',
    download_url='http://github.com/j-bennet/wharfee',
    author_email='i[dot]chernyavska[at]gmail[dot]com',
    version=__version__,
    license='LICENSE.txt',
    install_requires=[
        'six>=1.9.0',
        'pygments>=2.0.2',
        'prompt_toolkit>=1.0.0,<1.1.0',
        'docker-py>=1.6.0',
        'tabulate>=0.7.5',
        'click>=4.0',
        'py-pretty>=0.1',
        'configobj>=5.0.6',
        'pexpect>=3.3',
        'fuzzyfinder>=1.0.0'
    ],
    extras_require={
        'testing': [
            'pytest>=2.7.0',
            'mock>=1.0.1',
            'tox>=1.9.2'
        ],
    },
    entry_points={
        'console_scripts': [
            'wharfee = wharfee.main:cli',
            'wharfee-ops = scripts.optionizer:main',
        ]
    },
    packages=['wharfee'],
    package_data={'wharfee': ['wharfeerc']},
    scripts=[],
    name='wharfee',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8
from __future__ import unicode_literals
from __future__ import print_function

from prompt_toolkit.key_binding.manager import KeyBindingManager
from prompt_toolkit.keys import Keys


def get_key_manager(set_long_options, get_long_options, set_fuzzy_match, get_fuzzy_match):
    """
    Create and initialize keybinding manager
    :return: KeyBindingManager
    """

    assert callable(set_long_options)
    assert callable(get_long_options)
    assert callable(set_fuzzy_match)
    assert callable(get_fuzzy_match)

    manager = KeyBindingManager(
        enable_search=True,
        enable_system_bindings=True,
        enable_abort_and_exit_bindings=True)

    @manager.registry.add_binding(Keys.F2)
    def _(event):
        """
        When F2 has been pressed, fill in the "help" command.
        """
        event.cli.current_buffer.insert_text("help")

    @manager.registry.add_binding(Keys.F3)
    def _(_):
        """
        Enable/Disable long option name suggestion.
        """
        set_long_options(not get_long_options())

    @manager.registry.add_binding(Keys.F4)
    def _(_):
        """
        Enable/Disable fuzzy matching.
        """
        set_fuzzy_match(not get_fuzzy_match())

    @manager.registry.add_binding(Keys.F10)
    def _(event):
        """
        When F10 has been pressed, quit.
        """
        # Unused parameters for linter.
        raise EOFError

    @manager.registry.add_binding(Keys.ControlSpace)
    def _(event):
        """
        Initialize autocompletion at cursor.

        If the autocompletion menu is not showing, display it with the
        appropriate completions for the context.

        If the menu is showing, select the next completion.
        """
        b = event.cli.current_buffer
        if b.complete_state:
            b.complete_next()
        else:
            event.cli.start_completion(select_first=False)

    return manager



# -*- coding: utf-8
from pygments.lexer import RegexLexer
from pygments.lexer import words
from pygments.token import Operator, Keyword, Text

from .options import COMMAND_NAMES, all_option_names


class CommandLexer(RegexLexer):
    name = 'Command Line'
    aliases = ['cli']
    filenames = []

    tokens = {
        'root': [
            (words(tuple(COMMAND_NAMES), prefix=r'^', suffix=r'\b'),
             Operator.Word),
            (words(tuple(all_option_names()), prefix=r'', suffix=r'\b'),
             Keyword),
            (r'.*\n', Text),
        ]
    }
